import itertools
import numpy as np
# from pyjavaproperties import Properties
from vaex.ext import jprops
import collections
import functools
from vaex.ui.qt import *
import vaex.dataset
import vaex.ui.plot_windows
import logging as logging
import vaex.ui.qt as dialogs
import vaex.execution
import vaex.kld
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
# from matplotlib.backends.backend_qt4agg import NavigationToolbar2QTAgg as NavigationToolbar
from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT as NavigationToolbar
import matplotlib.pyplot as plt
import scipy.stats


logger = logging.getLogger("vaex.ranking")


# since we do many columns at once, a smallar buffer will lead to more resposiveness in the gui
buffer_size = 1e6


def unique_column_names(dataset):
    # return list(set(dataset.column_names) | set(dataset.virtual_columns.keys()))
    return dataset.get_column_names(virtual=True)


testing = False


class PlotDialog(QtGui.QDialog):
    def __init__(self, parent, width=5, height=4, dpi=100, **options):
        super(PlotDialog, self).__init__()
        self.parent_widget = parent
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.toolbar = NavigationToolbar(self.canvas, self)

        self.layout = QtGui.QVBoxLayout()
        self.layout.addWidget(self.toolbar)
        self.layout.addWidget(self.canvas)
        self.setLayout(self.layout)
        self.figure.canvas.mpl_connect('motion_notify_event', self._on_mouse_motion)
        self.figure.canvas.mpl_connect('button_press_event', self._on_mouse_button)

        # self.plot()

    def _on_mouse_motion(self, event):
        if event.xdata is not None and event.ydata is not None:
            self.on_mouse_motion(event)

    def _on_mouse_button(self, event):
        if event.xdata is not None and event.ydata is not None:
            self.on_mouse_button(event)

    def on_mouse_button(self, event):
        pass

    def on_mouse_motion(self, event):
        label, x, y = self.get_tooltip(event)
        self.set_tooltip(x, y, label)

    def set_tooltip(self, x, y, label):
        y = self.canvas.height() - 1 - y
        print(("motion", label, x, y))
        if label:
            print(("self.canvas.x/y()", self.canvas.geometry().x(), self.canvas.geometry().y()))
            print(("self.pos.x/y()", self.canvas.pos().x() + self.pos().x(), self.canvas.pos().y() + self.pos().y()))
            point = QtCore.QPoint(x + self.canvas.x() + self.pos().x(), y + self.canvas.y() + self.pos().y())
            QtGui.QToolTip.showText(point, label)

    def plot(self):
        pass


class RankPlot(PlotDialog):
    def __init__(self, parent, table):
        super(RankPlot, self).__init__(parent)
        self.table = table

        self.button_update_data = QtGui.QPushButton('Update data')
        self.button_update_data.clicked.connect(self.update_data)
        self.layout.addWidget(self.button_update_data)

        grid_layout = QtGui.QGridLayout()
        grid_layout.setColumnStretch(2, 1)
        grid_layout.setAlignment(QtCore.Qt.AlignTop)
        grid_layout.setSpacing(0)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addLayout(grid_layout)

        row = 0
        self.options = ["mutual_information", "rank(mutual_information)", "correlation_coefficient", "rank(correlation_coefficient)", "abs(correlation_coefficient)", "rank(abs(correlation_coefficient))"]
        self.expression_x = "mutual_information"
        self.expression_y = "correlation_coefficient"
        self.codeline_x = Codeline(self, "x", self.options,
                                   getter=attrgetter(self, "expression_x"),
                                   setter=attrsetter(self, "expression_x"),
                                   update=self.plot)
        self.codeline_y = Codeline(self, "y", self.options,
                                   getter=attrgetter(self, "expression_y"),
                                   setter=attrsetter(self, "expression_y"),
                                   update=self.plot)
        row = self.codeline_x.add_to_grid_layout(row, grid_layout)
        row = self.codeline_y.add_to_grid_layout(row, grid_layout)

        self.axes = self.figure.add_subplot(111)
        if 0:
            def onpick(event):
                thisline = event.artist
                xdata = thisline.get_xdata()
                ydata = thisline.get_ydata()
                index = event.ind
                print((index, xdata[index], ydata[index], self.pairs[index]))
                self.table.select_pair(self.pairs[index])
            self.figure.canvas.mpl_connect('pick_event', onpick)

        self.update_data()
        self.plot()

    def on_mouse_button(self, event):
        min_distance_index, x, y = self.find_nearest(event)
        print(("click", x, y))
        self.table.select_pair(self.pairs[min_distance_index])

        def update_tooltip():
            label, x, y = self.get_tooltip(event)
            self.set_tooltip(x, y, label)
        QtCore.QTimer.singleShot(100, update_tooltip)

    def find_nearest(self, event):
        transform = event.inaxes.transData.transform
        xy = transform(list(zip(self.x, self.y)))
        # print xy
        x, y = xy.T
        print(("event.x/y", event.x, event.y))
        distances = np.sqrt((x - event.x)**2 + (y - event.y)**2)
        min_distance_index = np.argmin(distances)
        return min_distance_index, x[min_distance_index], y[min_distance_index]

    def get_tooltip(self, event):
        min_distance_index, x, y = self.find_nearest(event)
        label = "-".join(self.pairs[min_distance_index])
        return label, x, y

    def update_data(self):
        self.variables = {}
        self.pairs = pairs = self.table.getSelected()
        mi = [self.table.qualities[pair] for pair in pairs]
        corr = [self.table.correlation_map[pair] for pair in pairs]
        self.variables["mutual_information"] = np.array(mi)
        self.variables["correlation_coefficient"] = np.array(corr)

    def plot(self):
        scope = {}
        scope.update(np.__dict__)
        scope.update(self.variables)
        scope["rank"] = scipy.stats.rankdata
        self.x = x = eval(self.expression_x, scope)
        self.y = y = eval(self.expression_y, scope)
        self.axes.cla()
        self.axes.plot(x, y, '.', picker=5)
        self.canvas.draw()


class ____RankingTableModel(QtCore.QAbstractTableModel):
    def __init__(self, dataset, dim=1, parent=None, *args):
        QtCore.QAbstractTableModel.__init__(self, parent, *args)
        self.dataset = dataset

        self.pairs = list(itertools.combinations(unique_column_names(self.dataset), dim))
        self.ranking = [None for pair in self.pairs]
        self.headers = ["subspace", "Mutual information", "MI ranking", "correlation", "MI ranking", 'selected']
        self.indices = list(range(len(self.pairs)))

    def rowCount(self, parent):
        return len(self.pairs)

    def columnCount(self, parent):
        return len(self.headers)

    def data(self, index, role):
        if not index.isValid():
            return None
        elif role != QtCore.Qt.DisplayRole:
            return None
        column = index.column()
        index = self.indices[index.row()]  # use sorted index
        if column == 0:
            return "-vs".join(self.pairs[index])
        if column == 1:
            rank = self.ranking[index]
            return "" if rank is None else str(rank)
        if column == 2:
            rank = self.ranking[index]
            return False if random.random() < 0.5 else True

    def headerData(self, index, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self.headers[index]
        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return str(index + 1)
        return None

    def sort(self, Ncol, order):
        """Sort table by given column number.
        """
        self.emit(QtCore.SIGNAL("layoutAboutToBeChanged()"))
        if Ncol == 0:
            print("by name")
            # get indices, sorted by pair name
            sortlist = list(zip(self.pairs, list(range(len(self.pairs)))))
            print(sortlist)
            sortlist.sort(key=operator.itemgetter(0))
            print(sortlist)
            self.indices = list(map(operator.itemgetter(1), sortlist))
            print((self.indices))
        if Ncol == 1:
            # get indices, sorted by ranking, or no sorting
            if None not in self.ranking:
                sortlist = list(zip(self.ranking, list(range(len(self.pairs)))))
                sortlist.sort(key=operator.itemgetter(0))
                self.indices = list(map(operator.itemgetter(1), sortlist))
            else:
                self.indices = list(range(len(self.pairs)))
            print((self.indices))
        if order == QtCore.Qt.DescendingOrder:
            self.indices.reverse()
        print((self.indices))
        self.emit(QtCore.SIGNAL("layoutChanged()"))


class SubspaceTable(QtGui.QTableWidget):
    def __init__(self, dialog, parent, mainPanel, dataset, pairs, dim, properties):
        self.dialog = dialog
        self.dim = dim
        if dim == 1:
            self.headers = ['', 'space', "min", "max", 'plot']
        else:
            self.headers = ['', 'space', "Mutual information", "Correlation", 'plot']
        self.properties = properties
        self.qualities = {}
        self.correlation_map = {}
        if testing:
            self.qualities = {key: np.random.random() for key in pairs}
            self.correlation_map = {key: np.random.normal() for key in pairs}

        # print ", ".join([""+("-".join(pair))+"" for pair in pairs])
        self.dataset = dataset
        self.filter_terms = []
        self.mainPanel = mainPanel
        self.pairs = list(pairs)  # list(itertools.combinations(self.dataset.column_names, dim))
        QtGui.QTableWidget.__init__(self, len(self.pairs), len(self.headers), parent)
        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.filter_mask = np.array([True for pair in pairs])
        self.selected_dict = {pair: self.properties.get(".".join(pair) + ".use", "True") == "True" for pair in pairs}
        # self.tableModel = RankingTableModel(self.dataset, dim, parent)
        # self.setModel(self.tableModel)
        # self.sortByColumn(0, QtCore.Qt.AscendingOrder)
        # self.setSortingEnabled(True)
        # self.pair_to_item = {}
        self.defaultFlags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsEditable
        # print self.properties._props
        if 1:
            # self.ranking = [None for pair in self.pairs]
            self.dim = dim
            self.setHorizontalHeaderLabels(self.headers)
            # self.setVerticalHeaderLabels(map(str, range(len(self.pairs))))
            self.fill_table()
            self.setSortingEnabled(True)
        self.queue_fill_table = vaex.ui.plot_windows.Queue("fill table", 200, self.fill_table)

    def pair_to_text(self, pair):
        return " ".join(map(str, pair))

    def select_pair(self, pair):
        # self.setSortingEnabled(False)
        index = self.pairs.index(pair)
        for i in range(self.rowCount()):
            item = self.item(i, 1)
            print((item.text(), self.pair_to_text(pair)))
            if item.text() == self.pair_to_text(pair):
                self.selectRow(i)
        # print index, self.visualRow(index)
        # self.selectRow(self.visualRow(index))
        # self.setSortingEnabled(True)

    def fill_table(self):
        # bug in qt? http://stackoverflow.com/questions/7960505/strange-qtablewidget-behavior-not-all-cells-populated-after-sorting-followed-b
        # fix: disable sorting, then enable again
        self.setSortingEnabled(False)
        self.checkboxes = []
        self.buttons = []
        pairs = [pair for pair, display in zip(self.pairs, self.filter_mask) if display]
        self.setRowCount(len(pairs))
        self.setVerticalHeaderLabels(list(map(str, list(range(len(pairs))))))
        for i in range(len(pairs)):
            pair = pairs[i]
            text = self.pair_to_text(pair)
            item = QtGui.QTableWidgetItem(text)
            self.setItem(i, 1, item)
            item.setFlags(self.defaultFlags)
            # item = QtGui.QTableWidgetItem()
            # item.setData(QtCore.Qt.DisplayRole, QtCore.QVariant(True))
            # item.setFlags(QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsUserCheckable)
            checkbox = QtGui.QCheckBox(self)
            use_key = ".".join(map(str, pair)) + ".use"
            # if self.dim == 1 and use_key in self.properties._props:
            # #print use_key, eval(self.properties[use_key])
            # checkbox.setCheckState(QtCore.Qt.Checked if eval(self.properties[use_key]) else QtCore.Qt.Unchecked)
            # else:
            # checkbox.setCheckState(QtCore.Qt.Checked)
            print(("fill", pair, self.selected_dict[pair]))
            checkbox.setCheckState(QtCore.Qt.Checked if self.selected_dict[pair] else QtCore.Qt.Unchecked)
            self.checkboxes.append(checkbox)
            self.setCellWidget(i, 0, checkbox)

            def stateChanged(state, pair=pair):
                self.selected_dict[pair] = state == QtCore.Qt.Checked
                print(("set", pair, "to", self.selected_dict[pair]))
            checkbox.stateChanged.connect(stateChanged)

            if self.dim == 1:
                button = QtGui.QPushButton("plot: " + text, self)

                def plot(_ignore=None, pair=pair):
                    ranges = [self.dialog.range_map[k] for k in pair]
                    self.mainPanel.histogram(*pair, ranges=ranges)
                button.clicked.connect(plot)
                self.setCellWidget(i, 4, button)

                min_key = pair[0] + ".min"
                max_key = pair[0] + ".max"
                if 1:
                    # print "test", min_key
                    if min_key in self.properties:
                        item = QtGui.QTableWidgetItem()  # "%s"  % quality)
                        value = self.properties[min_key]
                        # print "it is in... and value =", value
                        item.setText("%s" % value)
                        item.setData(QtCore.Qt.DisplayRole, float(value))
                        item.setFlags(self.defaultFlags)
                        self.setItem(i, 2, item)

                    if max_key in self.properties:
                        value = self.properties[max_key]
                        item = QtGui.QTableWidgetItem()  # "%s"  % quality)
                        item.setText("%s" % value)
                        item.setData(QtCore.Qt.DisplayRole, float(value))
                        item.setFlags(self.defaultFlags)
                        self.setItem(i, 3, item)
            else:
                # print "quality", quality, qualities
                # row = self.pairs.index(pair)
                quality = self.qualities.get(pair)
                if quality is not None:
                    item = QtGui.QTableWidgetItem()  # "%s"  % quality)
                    item.setText("%s" % quality)
                    item.setData(QtCore.Qt.DisplayRole, float(quality))
                    item.setFlags(self.defaultFlags)
                    self.setItem(i, 2, item)
                correlation = self.correlation_map.get(pair)
                if correlation is not None:
                    item = QtGui.QTableWidgetItem()  # "%s"  % quality)
                    item.setText("%s" % correlation)
                    item.setData(QtCore.Qt.DisplayRole, float(correlation))
                    item.setFlags(self.defaultFlags)
                    self.setItem(i, 3, item)

            if self.dim == 2:
                button = QtGui.QPushButton("plot: " + text, self)

                def plot(_ignore=None, pair=pair):
                    ranges = [self.dialog.range_map[k] for k in pair]
                    self.mainPanel.plotxy(*pair, ranges=ranges)
                button.clicked.connect(
                    plot)
                self.setCellWidget(i, 4, button)
                self.buttons.append(button)  # keep ref count
            if self.dim == 3:
                button = QtGui.QPushButton("plot: " + text, self)

                def plot(_ignore=None, pair=pair):
                    ranges = [self.dialog.range_map[k] for k in pair]
                    self.mainPanel.plotxyz(*pair, ranges=ranges)
                button.clicked.connect(plot)
                self.setCellWidget(i, 4, button)
                self.buttons.append(button)  # keep ref count
            # self.setItem(i, 1, item)
        self.setSortingEnabled(True)

    def getSelected(self):
        selection = [checkbox.checkState() == QtCore.Qt.Checked for checkbox in self.checkboxes]
        selected_pairs = [pair for pair, selected in zip(self.pairs, selection) if selected]
        return selected_pairs

    def setQualities(self, pairs, qualities):
        self.qualities = {}
        for quality, pair in zip(qualities, pairs):
            self.qualities[pair] = quality
            # item = self.pair_to_item[pair]
            # print "quality", quality, qualities
            # row = self.pairs.index(pair)
            # item = QtGui.QTableWidgetItem()#"%s"  % quality)
            # item.setText("%s"  % quality)
            # item.setData(QtCore.Qt.DisplayRole, float(quality))
            # item.setFlags(self.defaultFlags)
            # self.setItem(row, 2, item)
        self.fill_table()

    def set_correlations(self, correlation_map):
        self.correlation_map = dict(correlation_map)
        self.fill_table()

    def get_range(self, pair):
        index = self.pairs.index(pair)
        mi = self.item(index, 2)
        ma = self.item(index, 3)
        if mi is None or ma is None:
            return None, None
        # print pair, mi, ma
        # print mi.data(QtCore.Qt.DisplayRole)
        mi = None if mi is None else float(mi.data(QtCore.Qt.DisplayRole))
        ma = None if ma is None else float(ma.data(QtCore.Qt.DisplayRole))
        # print "->", pair, mi, ma
        return mi, ma

    def setRanges(self, pairs, ranges):
        for (mi, ma), pair in zip(ranges, pairs):
            # item = self.pair_to_item[pair]
            row = self.pairs.index(pair)
            item = QtGui.QTableWidgetItem()  # "%s"  % quality)
            item.setText("%s" % mi)
            item.setData(QtCore.Qt.DisplayRole, float(mi))
            item.setFlags(self.defaultFlags)
            self.setItem(row, 2, item)
            item = QtGui.QTableWidgetItem()  # "%s"  % quality)
            item.setText("%s" % ma)
            item.setData(QtCore.Qt.DisplayRole, float(ma))
            item.setFlags(self.defaultFlags)
            self.setItem(row, 3, item)

    def deselect(self, pair):
        index = self.pairs.index(pair)
        print(("deselect", pair, index))
        checkbox = self.checkboxes[index]
        checkbox.setCheckState(QtCore.Qt.Unchecked)

    def select(self, pair):
        index = self.pairs.index(pair)
        print(("deselect", pair, index))
        checkbox = self.checkboxes[index]
        checkbox.setCheckState(QtCore.Qt.Checked)

    def setPairs(self, pairs):
        # selection = [checkbox.checkState() == QtCore.Qt.Checked for checkbox in self.checkboxes]
        # non_selected_pairs = [pair for pair, selected in zip(self.pairs, selection) if not selected]

        self.pairs = list(pairs)
        for pair in self.pairs:
            if pair not in self.selected_dict:
                self.selected_dict[pair] = self.properties.get(".".join(pair) + ".use", "True")
        self.filter_mask = np.array([True for pair in pairs])
        self.fill_table()
        # self.checkboxes = []
        # self.setRowCount(len(self.pairs))
        # self.setVerticalHeaderLabels(map(str, range(len(self.pairs))))
        # for i in range(len(self.pairs)):
        # text = " ".join(map(str, self.pairs[i]))
        # print text
        # item = QtGui.QTableWidgetItem(text)
        # item.setFlags(self.defaultFlags)
        # self.setItem(i, 1, item)
        # checkbox = self.checkboxes[i] #QtGui.QCheckBox(self)
        # if not (self.pairs[i] in non_selected_pairs):
        # checkbox.setCheckState(QtCore.Qt.Checked)
        # self.checkboxes.append(checkbox)
        # self.setCellWidget(i, 0, checkbox)
        # print self.checkboxes

    def set_filter_terms(self, filter_terms):
        def filter(pair):
            found = True
            for filter_term in filter_terms:
                found_term = False
                for expression in pair:
                    found_term = found_term or filter_term.lower() in expression.lower()
                found = found and found_term
            return found

        self.filter_terms = filter_terms
        # print list, filter, self.pairs
        self.filter_mask = np.array([filter(pair) for pair in self.pairs])
        self.queue_fill_table()
        # self.fill_table()


def joinpairs(pairs1d, pairsNd):
    previous = []
    for pair1d in pairs1d:
        subspacename = pair1d[0]  # tuple only has one element
        for pairNd in pairsNd:
            if subspacename not in pairNd:
                pair = pair1d + pairNd
                if sorted(pair) not in previous:
                    previous.append(sorted(pair))
                    # print previous
                    yield pair


class RankDialog(QtGui.QDialog):
    def __init__(self, dataset, parent, mainPanel, **options):
        super(RankDialog, self).__init__(parent)
        self.dataset = dataset
        self.mainPanel = mainPanel
        self.range_map = {}
        self.grid_size = int(options.get("grid_size", "32"))

        # print "options", options
        self.properties = collections.OrderedDict()  # Properties()
        if self.dataset.is_local():
            self.properties_path = os.path.splitext(self.dataset.path)[0] + ".properties"
        else:
            dir = os.path.join(vaex.utils.get_private_dir(), "ranking")
            if not os.path.exists(dir):
                os.makedirs(dir)
            server = self.dataset.server
            name = "%s_%s_%s_%s" % (server.hostname, server.port, server.base_path.replace("/", "_"), self.dataset.name)
            self.properties_path = os.path.join(dir, name + ".properties")

        self.properties_path = options.get("file", self.properties_path)
        if os.path.exists(self.properties_path):
            self.load_properties()
        else:
            pass
            # if not os.access(properties_path, os.W_OK):
            # dialog_error(self, "File access", "Cannot write to %r, so cannot save options" % properties_path)

        self.tabs = QtGui.QTabWidget(self)

        self.tab1d = QtGui.QWidget(self.tabs)
        self.table1d = SubspaceTable(self, self.tab1d, mainPanel, self.dataset, list(itertools.combinations(unique_column_names(self.dataset), 1)), 1, self.properties)

        self.subspaceTables = {}
        self.subspaceTabs = {}
        self.subspaceTables[1] = self.table1d
        self.subspaceTabs[1] = self.tab1d

        def onclick(dim=2):
            self.open(dim=dim)
        self.subspace2d = QtGui.QPushButton("create 2d subspaces", self.tab1d)
        self.subspace2d.clicked.connect(functools.partial(onclick, dim=2))

        self.get_ranges_menu = QtGui.QMenu()
        self.button_get_ranges = QtGui.QToolButton()
        self.button_get_ranges.setText("calculate min/max")
        self.button_get_ranges.setPopupMode(QtGui.QToolButton.InstantPopup)
        # self.button_get_ranges.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        self.get_ranges_menu = QtGui.QMenu()
        self.button_get_ranges.setMenu(self.get_ranges_menu)

        self.action_ranges_minmax = QtGui.QAction("absolute min/max", self)
        self.action_ranges_minmax_3sigma = QtGui.QAction("3 sigma clipping", self)
        self.get_ranges_menu.addAction(self.action_ranges_minmax)
        self.get_ranges_menu.addAction(self.action_ranges_minmax_3sigma)

        # self.button_get_ranges = QtGui.QToolButton(self.tab1d)
        # self.button_get_ranges.setText("calculate min/max")
        # self.button_get_ranges.setM
        self.action_ranges_minmax.triggered.connect(self.onCalculateMinMax)
        self.action_ranges_minmax_3sigma.triggered.connect(self.onCalculateMinMax3Sigma)

        self.button_store = QtGui.QToolButton(self.tab1d)
        self.button_store.setText("store")
        self.button_store.clicked.connect(self.onStore)

        self.actions_menu = QtGui.QMenu()
        self.button_actions = QtGui.QToolButton()
        self.button_actions.setText("Extra")
        self.button_actions.setPopupMode(QtGui.QToolButton.InstantPopup)
        self.button_actions.setMenu(self.actions_menu)

        self.action_select_all = QtGui.QAction("Select all", self)
        self.action_select_none = QtGui.QAction("Select none", self)
        self.action_remove_empty = QtGui.QAction("Remove empty columns", self)
        self.action_pca = QtGui.QAction("PCA transformation", self)
        self.action_select_all.triggered.connect(self.onSelectAll)
        self.action_select_none.triggered.connect(self.onSelectNone)
        self.action_remove_empty.triggered.connect(self.onRemoveEmpty)
        self.action_pca.triggered.connect(self.onPca)
        self.actions_menu.addAction(self.action_select_all)
        self.actions_menu.addAction(self.action_select_none)
        self.actions_menu.addAction(self.action_remove_empty)
        self.actions_menu.addSeparator()
        self.actions_menu.addAction(self.action_pca)

        self.tab1dlayout = QtGui.QVBoxLayout(self)
        self.tab1d_button_layout = QtGui.QHBoxLayout(self)
        self.tab1dlayout.addLayout(self.tab1d_button_layout)
        self.tab1d_button_layout.addWidget(self.subspace2d)
        self.tab1d_button_layout.addWidget(self.button_get_ranges)
        self.tab1d_button_layout.addWidget(self.button_store)
        self.tab1d_button_layout.addWidget(self.button_actions)

        self.filter_line_edit = QtGui.QLineEdit(self)
        self.filter_line_edit.setPlaceholderText("Enter space seperated search terms")
        self.filter_line_edit.textEdited.connect(functools.partial(self.onFilter, table=self.table1d))

        self.tab1dlayout.addWidget(self.filter_line_edit)
        self.tab1dlayout.addWidget(self.table1d)
        # self.tab1dlayout.addWidget(self.rankButton)
        # self.setCentralWidget(self.splitter)
        self.tab1d.setLayout(self.tab1dlayout)

        self.tabs.addTab(self.tab1d, "1d")

        self.resize(700, 500)

        if 0:
            for name in self.dataset.column_names:
                item = QtGui.QListWidgetItem(self.list1d)
                item.setText(name)
                item.setCheckState(False)
                # self.list1d.

        self.boxlayout = QtGui.QVBoxLayout(self)

        self.gridlayout = QtGui.QGridLayout()
        self.gridlayout.setColumnStretch(1, 1)
        self.gridlayout.setSpacing(0)
        self.gridlayout.setContentsMargins(2, 1, 2, 1)
        self.gridlayout.setAlignment(QtCore.Qt.AlignTop)
        row = 1

        self.selection_label = QtGui.QLabel("Use for computations:", self)
        self.gridlayout.addWidget(self.selection_label, row, 0)
        self.radio_button_all = QtGui.QRadioButton("Complete dataset", self)
        self.radio_button_selection = QtGui.QRadioButton("Selection", self)
        self.radio_button_all.setChecked(True)
        self.gridlayout.addWidget(self.radio_button_all, row, 1)
        row += 1
        self.gridlayout.addWidget(self.radio_button_selection, row, 1)
        row += 1

        def get():
            return str(self.grid_size)

        def set(value):
            self.grid_size = int(value)
        self.option_grid_size = Option(self, "grid size (for mutual info)", "32 64 128 256".split(), get, set)
        row = self.option_grid_size.add_to_grid_layout(row, self.gridlayout)
        # self.gridlayout.addWidget(self.option_grid_size.combobox); row += 1
        # self.gridlayout.addWidget(self.option_grid_size.combobox); row += 1

        self.boxlayout.addLayout(self.gridlayout)
        self.boxlayout.addWidget(self.tabs)
        # self.boxlayout.addWidget(self.rankButton)
        # self.setCentralWidget(self.splitter)
        self.setLayout(self.boxlayout)

        if "2" in options.get("open", ""):
            self.open(dim=2)
        if "3" in options.get("open", ""):
            self.open(dim=3)
        if "4" in options.get("open", ""):
            self.open(dim=4)

        self.fill_range_map()

    def open(self, dim=2):
        pairs1d = self.subspaceTables[1].getSelected()
        pairsprevd = self.subspaceTables[dim - 1].getSelected()
        # print pairs1d
        # print pairsprevd
        newpairs = list(joinpairs(pairs1d, pairsprevd))
        print(("newpairs", newpairs))
        if dim not in self.subspaceTables:
            self.tabNd = QtGui.QWidget(self.tabs)
            self.tableNd = SubspaceTable(self, self.tabNd, self.mainPanel, self.dataset, newpairs, dim, self.properties)
            self.tabNdlayout = QtGui.QVBoxLayout(self)
            self.tabNdButtonLayout = QtGui.QHBoxLayout(self)
            self.subspaceNd = QtGui.QPushButton("Create %dd subspaces" % (dim + 1), self.tab1d)
            self.plotNd = QtGui.QPushButton("Rank plot")
            self.exportNd = QtGui.QPushButton("Export ranking")
            if dim == len(self.dataset.column_names):
                self.subspaceNd.setDisabled(True)

            self.menu_calculate = QtGui.QMenu()
            self.button_calculate = QtGui.QToolButton()
            self.button_calculate.setText("Calculate")
            self.button_calculate.setPopupMode(QtGui.QToolButton.InstantPopup)
            # self.button_get_ranges.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
            self.button_calculate.setMenu(self.menu_calculate)

            self.tabNdButtonLayout.addWidget(self.subspaceNd)
            self.tabNdButtonLayout.addWidget(self.button_calculate)
            # self.tabNdButtonLayout.addWidget(self.miNd)
            # self.tabNdButtonLayout.addWidget(self.correlationNd)
            self.tabNdButtonLayout.addWidget(self.exportNd)
            self.tabNdButtonLayout.addWidget(self.plotNd)
            self.tabNdlayout.addLayout(self.tabNdButtonLayout)
            self.subspaceNd.clicked.connect(functools.partial(self.open, dim=dim + 1))
            self.exportNd.clicked.connect(functools.partial(self.export, table=self.tableNd))
            self.plotNd.clicked.connect(functools.partial(self.rank_plot, table=self.tableNd))

            self.action_calculate_mi = QtGui.QAction("Calculate mutual information", self)
            self.action_calculate_correlation = QtGui.QAction("Calculate correlation", self)
            self.action_calculate_mi.triggered.connect(functools.partial(self.rankSubspaces, table=self.tableNd))
            self.action_calculate_correlation.triggered.connect(functools.partial(self.calculate_correlation, table=self.tableNd))
            self.menu_calculate.addAction(self.action_calculate_mi)
            self.menu_calculate.addAction(self.action_calculate_correlation)

            self.action_calculate_rank_correlation_MI_corr_kendall = QtGui.QAction("MI - correlation", self)
            self.action_calculate_rank_correlation_MI_abs_corr_kendall = QtGui.QAction("MI - abs(correlation)", self)

            self.action_calculate_rank_correlation_MI_corr_spearman = QtGui.QAction("MI - correlation", self)
            self.action_calculate_rank_correlation_MI_abs_corr_spearman = QtGui.QAction("MI - abs(correlation)", self)

            self.menu_correlation_kendall = self.menu_calculate.addMenu("Kendall's rank correlation")
            self.menu_correlation_spearman = self.menu_calculate.addMenu("Spearman's rank correlation")
            self.menu_correlation_kendall.setEnabled(False)
            self.menu_correlation_spearman.setEnabled(False)

            self.menu_correlation_kendall.addAction(self.action_calculate_rank_correlation_MI_corr_kendall)
            self.menu_correlation_kendall.addAction(self.action_calculate_rank_correlation_MI_abs_corr_kendall)

            self.menu_correlation_spearman.addAction(self.action_calculate_rank_correlation_MI_corr_spearman)
            self.menu_correlation_spearman.addAction(self.action_calculate_rank_correlation_MI_abs_corr_spearman)

            self.action_calculate_rank_correlation_MI_corr_kendall.triggered.connect(functools.partial(self.calculate_rank_correlation_kendall, table=self.tableNd))
            self.action_calculate_rank_correlation_MI_abs_corr_kendall.triggered.connect(functools.partial(self.calculate_rank_correlation_kendall, table=self.tableNd, absolute=True))

            self.action_calculate_rank_correlation_MI_corr_spearman.triggered.connect(functools.partial(self.calculate_rank_correlation_spearman, table=self.tableNd))
            self.action_calculate_rank_correlation_MI_abs_corr_spearman.triggered.connect(functools.partial(self.calculate_rank_correlation_spearman, table=self.tableNd, absolute=True))

            def func(index, name=""):
                print((name, index.row(), index.column()))
            self.tableNd.pressed.connect(functools.partial(func, name="pressed"))
            self.tableNd.entered.connect(functools.partial(func, name="entered"))
            self.tableNd.clicked.connect(functools.partial(func, name="clicked"))
            self.tableNd.activated.connect(functools.partial(func, name="activated"))

            def func(index, previous, name=""):
                print((name, index.row(), index.column(), previous.row(), previous.column()))
            self.selectionModel = self.tableNd.selectionModel()
            self.selectionModel.currentChanged.connect(functools.partial(func, name="currentChanged"))

            self.filter_Nd_line_edit = QtGui.QLineEdit(self)
            self.filter_Nd_line_edit.setPlaceholderText("Enter space seperated search terms")
            self.filter_Nd_line_edit.textEdited.connect(functools.partial(self.onFilter, table=self.tableNd))
            self.tabNdlayout.addWidget(self.filter_Nd_line_edit)

            self.tabNdlayout.addWidget(self.tableNd)
            # self.tab1dlayout.addWidget(self.rankButton)
            # self.setCentralWidget(self.splitter)
            self.tabNd.setLayout(self.tabNdlayout)
            self.subspaceTables[dim] = self.tableNd
            self.subspaceTabs[dim] = self.tabNd

            self.tabs.addTab(self.tabNd, "%dd" % dim)
            self.tabs.setCurrentWidget(self.tabNd)
        else:
            self.subspaceTables[dim].setPairs(newpairs)
            self.tabs.setCurrentWidget(self.subspaceTabs[dim])

    def onPca(self):
        # vaex.pca.
        pass

    def onFilter(self, text, table):
        table.set_filter_terms(text.split())

    def onStore(self):
        selected_pairs = self.table1d.getSelected()
        # error = False
        for pair in self.table1d.pairs:
            key = str(pair[0])
            # print repr(key+".use"), repr(pair in selected_pairs)
            self.properties[key + ".use"] = repr(pair in selected_pairs)
            if key in self.range_map:
                mi, ma = self.range_map[key]
                self.properties[key + ".min"] = repr(mi)
                self.properties[key + ".max"] = repr(ma)
            else:
                print(("min/max not present", key))
        print(("save to", self.properties_path))
        self.store_properties()
        dialog_info(self, "Stored", "Stored configuration to: %r" % self.properties_path)

    def load_properties(self):
        with open(self.properties_path, "rb") as f:
            self.properties = jprops.load_properties(f, collections.OrderedDict)

    def store_properties(self):
        with open(self.properties_path, "wb") as f:
            jprops.store_properties(f, self.properties)

    def fill_range_map(self):
        pairs = self.table1d.getSelected()
        for pair in pairs:
            mi, ma = self.table1d.get_range(pair)
            if mi is not None and ma is not None:
                self.range_map[pair[0]] = mi, ma

    def onSelectAll(self):
        pairs = self.table1d.pairs
        for pair in pairs:
            self.table1d.select(pair)

    def onSelectNone(self):
        pairs = self.table1d.pairs
        for pair in pairs:
            self.table1d.deselect(pair)

    def onRemoveEmpty(self):
        self.fill_range_map()
        pairs = self.table1d.getSelected()
        error = False
        for pair in pairs:
            print(pair)
            if pair[0] in self.range_map:
                min, max = self.range_map[pair[0]]
                if min == max:
                    print((pair, "empty", min, max))
                    self.table1d.deselect(pair)
            else:
                if not error:  # only give a warning once
                    dialog_error(self, "Min/max missing", "Min max missing for %s" % pair)
                    error = True

    def onCalculateMinMax(self):
        pairs = self.table1d.getSelected()
        logger.debug("estimate min/max for %r" % pairs)
        if self.dataset.is_local():
            executor = vaex.execution.Executor(buffer_size=buffer_size)
        else:
            executor = vaex.remote.ServerExecutor()

        expressions = [pair[0] for pair in pairs]
        assert len(pairs[0]) == 1
        self.range_map = {}
        try:
            with dialogs.ProgressExecution(self, "Calculating min/max", executor=executor) as progress:
                subspace = self.dataset.subspace(*expressions, executor=executor, delay=True)
                minmax = subspace.minmax()
                progress.add_task(minmax).end()
                progress.execute()
            ranges = minmax.get()
            self.table1d.setRanges(pairs, ranges)
            self.fill_range_map()
        except:
            logger.exception("Error in min/max or cancelled")
        # dialog.hide()

    def onCalculateMinMax3Sigma(self):
        pairs = self.table1d.getSelected()

        expressions = [pair[0] for pair in pairs]
        if self.dataset.is_local():
            executor = vaex.execution.Executor(buffer_size=buffer_size)
        else:
            executor = vaex.remote.ServerExecutor()

        if self.dataset.is_local():
            executor = vaex.execution.Executor()
        else:
            executor = vaex.remote.ServerExecutor()
        subspace = self.dataset.subspace(*expressions, executor=executor, delay=True)
        means = subspace.mean()
        with dialogs.ProgressExecution(self, "Calculating mean", executor=executor) as progress:
            progress.add_task(means).end()
            progress.execute()
        logger.debug("get means")
        means = means.get()
        logger.debug("got means")

        vars = subspace.var(means=means)
        with dialogs.ProgressExecution(self, "Calculating variance", executor=executor) as progress:
            progress.add_task(vars).end()
            progress.execute()
        # limits  = limits.get()
        vars = vars.get()
        stds = vars**0.5
        sigmas = 3
        ranges = list(zip(means - sigmas * stds, means + sigmas * stds))
        self.table1d.setRanges(pairs, ranges)
        self.fill_range_map()

    def calculate_rank_correlation_kendall(self, table, absolute=False):
        print(("kendall", table, absolute))
        pairs = table.getSelected()
        mi_values = [table.qualities[pair] for pair in pairs]
        correlation_values = [table.correlation_map[pair] for pair in pairs]
        ranking_mi = np.argsort(mi_values)
        if absolute:
            ranking_correlation = np.argsort(np.abs(correlation_values))
        else:
            ranking_correlation = np.argsort(correlation_values)
        N = len(pairs)
        A = np.zeros((N, N))
        B = np.zeros((N, N))
        for i in range(N):
            for j in range(N):
                A[i, j] = np.sign(ranking_mi[i] - ranking_mi[j])
                B[i, j] = np.sign(ranking_correlation[i] - ranking_correlation[j])
        AB = 0
        AA = 0
        BB = 0
        for i in range(N):
            for j in range(N):
                AB += A[i, j] * B[i, j]
                AA += A[i, j]**2
                BB += B[i, j]**2

    def calculate_rank_correlation_spearman(self, table, absolute=False):
        print(("spearman", table, absolute))

    def calculate_correlation(self, table):
        print(("calculate correlation for ", table))
        pairs = table.getSelected()

        expressions = set()
        for pair in pairs:
            for expression in pair:
                expressions.add(expression)
        expressions = list(expressions)
        if self.dataset.is_local():
            executor = vaex.execution.Executor(buffer_size=buffer_size)
        else:
            executor = vaex.remote.ServerExecutor()

        def on_error(exc):
            raise exc
        if 1:
            # subspace = self.dataset(*expressions, executor=executor, delay=True)
            subspaces = self.dataset.subspaces(pairs, executor=executor, delay=True)
            means_promise = subspaces.mean()
            # print means_promise, type(means_promise), subs
            with dialogs.ProgressExecution(self, "Calculating means", executor=executor) as progress:
                progress.add_task(means_promise)
                progress.execute()
            means = means_promise.get()

            variances_promise = subspaces.var(means=means)
            with dialogs.ProgressExecution(self, "Calculating variances", executor=executor) as progress:
                progress.add_task(variances_promise)
                progress.execute()
            vars = variances_promise.get()

            # means = subspaces._unpack(means_packed)
            # vars = subspaces._unpack(vars_packed)
            tasks = []
            with dialogs.ProgressExecution(self, "Calculating correlation", executor=executor) as progress:
                for subspace, mean, var in zip(subspaces.subspaces, means, vars):
                    task = subspace.correlation(means=mean, vars=var)
                    progress.add_task(task).end()
                    tasks.append(task)
                progress.execute()
            correlations = [task.get() for task in tasks]

            correlation_map = dict(zip(pairs, correlations))
            table.set_correlations(correlation_map)
            return
            # mean_map = dict(zip(expressions, means))
            # var_map = dict(zip(expressions, variances))
        else:
            mean_map = {}

            def on_error(exc):
                raise exc
            for expression in expressions:
                subspace = self.dataset(expression, executor=executor, delay=True)

                def assign(mean_list, expression=expression):
                    logger.debug("assigning %r to %s", mean_list, expression)
                    mean_map[expression] = mean_list
                subspace.mean().then(assign, on_error).end()
            with dialogs.ProgressExecution(self, "Calculating means", executor=executor):
                executor.execute()

            var_map = {}
            for expression in expressions:
                subspace = self.dataset(expression, executor=executor, delay=True)

                def assign(mean_list, expression=expression):
                    logger.debug("assigning %r to %s", mean_list, expression)
                    var_map[expression] = mean_list[0].tolist()
                subspace.var(means=mean_map[expression]).then(assign, on_error).end()
            with dialogs.ProgressExecution(self, "Calculating variances", executor=executor):
                executor.execute()

            means = [mean_map[expressions[0]] for expressions in pairs]
            variances = [var_map[expressions[0]] for expressions in pairs]

        correlation_map = {}
        for pair in pairs:
            means = [mean_map[expression] for expression in pair]
            vars = [var_map[expression] for expression in pair]
            subspace = self.dataset(*pair, executor=executor, delay=True)

            def assign(correlation, pair=pair):
                logger.debug("assigning %r to %s", correlation, pair)
                correlation_map[pair] = correlation
            subspace.correlation(means, vars).then(assign, on_error).end()

        with dialogs.ProgressExecution(self, "Calculating correlation", executor=executor):
            executor.execute()

        table.set_correlations(correlation_map)
        return

        jobsManager = vaex.dataset.JobsManager()
        expressions = set()
        for pair in pairs:
            for expression in pair:
                expressions.add(expression)
        expressions = list(expressions)
        print("means")
        with ProgressExecution(self, "Calculating means") as progress:
            means = jobsManager.calculate_mean(self.dataset, use_mask=self.radio_button_selection.isChecked(), expressions=expressions, feedback=progress.progress)
        mean_map = dict(list(zip(expressions, means)))
        centered_expressions_map = {expression: "(%s - %.20e)" % (expression, mean) for (expression, mean) in list(mean_map.items())}
        variances_expressions_map = {expression: "%s**2" % centered_expressions for expression, centered_expressions in list(centered_expressions_map.items())}
        with ProgressExecution(self, "Calculating variances") as progress:
            variances = jobsManager.calculate_mean(self.dataset, use_mask=self.radio_button_selection.isChecked(), expressions=list(variances_expressions_map.values()), feedback=progress.progress)
        variances_map = dict(list(zip(list(variances_expressions_map.keys()), variances)))

        covariances_expressions = []
        for pair in pairs:
            centered_expressions = [centered_expressions_map[expression] for expression in pair]
            covariance_expression = "*".join(centered_expressions)
            covariances_expressions.append(covariance_expression)

        print(covariances_expressions)
        with ProgressExecution(self, "Calculating covariances") as progress:
            # progress.progress(20)
            covariances = jobsManager.calculate_mean(self.dataset, use_mask=self.radio_button_selection.isChecked(), expressions=covariances_expressions, feedback=progress.progress)
            # progress.progress(20)
        print(variances)
        print(covariances)

        correlation_map = {}
        for pair, covariance in zip(pairs, covariances):
            normalization = 1
            for expression in pair:
                normalization *= np.sqrt(variances_map[expression])
            correlation_map[pair] = covariance / normalization
        table.set_correlations(correlation_map)

        return

    def export(self, table):
        print(("export", table))
        basename, ext = os.path.splitext(self.dataset.path)
        path = basename + ("-ranking-%dd" % table.dim) + ".properties"
        filename = get_path_save(self, path=path, title="Export ranking", file_mask="properties file *.properties")
        expressions = set()
        if filename:
            counts = 0
            with open(filename, "w") as file:
                for pair in table.pairs:
                    if pair in table.qualities:
                        file.write("%s.mutual_information=%f\n" % (".".join(pair), table.qualities[pair]))
                        counts += 1
                    if pair in table.correlation_map:
                        file.write("%s.correlation_coefficient=%f\n" % (".".join(pair), table.correlation_map[pair]))
                        counts += 1
                    expressions.update(pair)
                for expression in expressions:
                    mi, ma = self.range_map[expression]
                    file.write("%s.min=%f\n" % (expression, mi))
                    file.write("%s.max=%f\n" % (expression, ma))
                dialog_info(self, "Wrote ranking file", "wrote %d lines" % counts)

    def rankSubspaces(self, table):
        self.fill_range_map()

        pairs = table.getSelected()
        error = False
        ranges = []

        for pair in pairs:
            for expression in pair:
                if expression not in self.range_map:
                    error = True
                    print(("missing", expression))
        if error:
            dialog_error(self, "Missing min/max", "Please calculate the minimum and maximum for the dimensions")
            return

        # expressions = [pair[0] for pair in pairs]
        # executor = vaex.execution.Executor(buffer_size=buffer_size)
        if self.dataset.is_local():
            executor = vaex.execution.Executor(buffer_size=buffer_size)
        else:
            executor = vaex.remote.ServerExecutor()

        tasks = []
        with dialogs.ProgressExecution(self, "Calculating mutual information", executor=executor) as progress:
            for pair in pairs:
                limits = [self.range_map[expr] for expr in pair]
                task = self.dataset(*pair, executor=executor, delay=True).mutual_information(limits=limits, size=self.grid_size)
                progress.add_task(task).end()
                tasks.append(task)
            if not progress.execute():
                return
        logger.debug("get means")
        mutual_information = [task.get() for task in tasks]

        # mutual_information_list = [MI_map[pair] for pair in pairs]
        table.setQualities(pairs, mutual_information)
        return

        print(table)
        qualities = []
        pairs = table.getSelected()

        if 0:
            for pair in pairs:
                dim = len(pair)
                # if dim == 2:
                columns = [self.dataset.columns[name] for name in pair]
                print(pair)
                information = vaex.kld.kld_shuffled(columns, mask=mask)
                qualities.append(information)
                # print pair
        if 0:
            dialog = QtGui.QProgressDialog("Calculating Mutual information", "Abort", 0, 1000, self)
            dialog.show()

            def feedback(percentage):
                print(percentage)
                dialog.setValue(int(percentage * 10))
                QtCore.QCoreApplication.instance().processEvents()
                if dialog.wasCanceled():
                    return True
        with ProgressExecution(self, "Calculating Mutual information") as progress:
            qualities = vaex.kld.kld_shuffled_grouped(self.dataset, self.range_map, pairs, feedback=progress.progress, use_mask=self.radio_button_selection.isChecked())
            # dialog.hide()
        if qualities is not None:
            print(qualities)
            table.setQualities(pairs, qualities)

    def rank_plot(self, table):
        plot = RankPlot(self, table)
        plot.show()



__author__ = 'breddels'
import javaobj
import sys
import io


if __name__ == "__main__":
    import logging
    javaobj._log.setLevel(logging.DEBUG)
    jobj = file(sys.argv[1]).read()[16 + 5:]
    print((repr(jobj[:100])))
    pobj, index = javaobj.loads(jobj)
    rest = jobj[index:]
    import zlib
    print((repr(rest[:100])))
    datastr = zlib.decompress(rest, -15)
    stream = io.StringIO(datastr)
    m = javaobj.JavaObjectUnmarshaller(stream)
    data = m.readObject()
    while len(data) == 2:
        print((data[0].classdesc.name))
        # datastr = datastr[data[1]:]
        x = stream.read(3)
        print(("data left", [hex(ord(k)) for k in x]))
        stream.seek(-3, 1)
        # x = stream.read(10)
        # print "data left", [hex(ord(k)) for k in x]
        # stream.seek(-10, 1)
        # data = javaobj.loads(datastr)
        data = m.readObject()
        print(data)
        if data[0] == "END":
            print("end...")
            x = stream.read(10)
            i
            print(("data left", [hex(ord(k)) for k in x]))

    print(data)
    # print pobj, index



# -*- coding: utf-8 -*-
import math
from vaex.dataset import Dataset
from vaex.ui.qt import *
import logging

logger = logging.getLogger("vaex.table")


PAGE_LIMIT = int(1e7)


class FullTableModel(QtCore.QAbstractTableModel):
    def __init__(self, dataset, page_size, page, parent=None, *args):
        """
        :type dataset: Dataset
        """
        QtCore.QAbstractTableModel.__init__(self, parent, *args)
        self.dataset = dataset
        self.page_size = page_size
        self.page = page
        self.row_count_start = 1

    def rowCount(self, parent):
        # print self.dataset._length
        # return int(self.dataset._length)
        left_over = len(self.dataset) - self.page * self.page_size
        if left_over < 0:
            return 0
        elif left_over < self.page_size:
            return left_over
        else:
            return self.page_size

    def set_page(self, page):
        self.page = page

    def get_row_offset(self):
        return self.page * self.page_size

    def get_column_names(self):
        return self.dataset.get_column_names(virtual=True, hidden=True, strings=True)

    def columnCount(self, parent):
        # return self.dataset.column_count()+1
        return len(self.get_column_names()) + 1

    def data(self, index, role):
        # return ""
        row_offset = self.get_row_offset()
        if not index.isValid():
            return None
        elif role != QtCore.Qt.DisplayRole:
            return None
        if index.column() == 0:
            return "{:,}".format(index.row() + self.row_count_start + row_offset)
        else:
            # column = self.dataset.all_columns[self.dataset.all_column_names[index.column()-1]]
            row = row_offset + index.row()
            column_name = self.get_column_names()[index.column() - 1]
            # if column_name in self.dataset.columns.keys():
            # value = self.dataset.columns[column_name][row]
            # else:
            # value, = self.dataset(column_name).row(row)
            try:
                value = self.dataset.evaluate(column_name, row, row + 1)
            except Exception as e:
                logger.exception("Error evaluating: %s %s", column_name, row)
                return "Error: %r" % e
            try:
                value = value[0]
                if len(value.shape) == 0:
                    return str(value)
                else:
                    return "%s %s" % (value.dtype.name, value.shape)
            except:
                pass  # TODO: ok, it can be a scalar sometimes... should not happen, but can when an column expression is set to '0'
                return str(value)
            # column = [self.dataset.all_column_names[index.column()-1]]

    def headerData(self, index, orientation, role):
        # print index
        row_offset = self.get_row_offset()
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            if index == 0:
                return "row"
            else:
                return self.get_column_names()[index - 1]
        # if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
        # return str(index+self.row_count_start + row_offset)
        return None


class TableDialog(QtGui.QDialog):
    def __init__(self, dataset, parent):
        super(TableDialog, self).__init__(parent)
        self.dataset = dataset
        # self.setModal(False)
        # self.setWindowFlags(QtCore.Qt#.FramelessWindowHint)

        self.resize(700, 500)
        self.tableView = QtGui.QTableView()
        self.tableView.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.header = self.dataset.column_names
        self.tableModel = FullTableModel(self.dataset, PAGE_LIMIT, 0, self)
        self.tableView.setModel(self.tableModel)
        self.tableView.pressed.connect(self.onSelectRow)

        if qt_mayor == 5:
            self.tableView.verticalHeader().setSectionResizeMode(QtGui.QHeaderView.Interactive)
        else:
            self.tableView.verticalHeader().setResizeMode(QtGui.QHeaderView.Interactive)

        if 0:
            for name in self.dataset.column_names:
                item = QtGui.QListWidgetItem(self.list1d)
                item.setText(name)
                item.setCheckState(False)
                # self.list1d.

        pages = int(math.ceil(len(self.dataset) * 1. / self.tableModel.page_size))

        self.boxlayout = QtGui.QVBoxLayout(self)
        self.header_layout = QtGui.QHBoxLayout(self)
        self.spinner = QtGui.QSpinBox(self)
        self.label_prefix = QtGui.QLabel("should not see me")
        self.label_postfix = QtGui.QLabel("should not see me")
        self.header_layout.addWidget(self.label_prefix)
        self.header_layout.addWidget(self.spinner)
        self.header_layout.addWidget(self.label_postfix)

        self.count_from_zero = QtGui.QCheckBox("count rows from zero", self)
        self.boxlayout.addWidget(self.count_from_zero)
        self.boxlayout.addLayout(self.header_layout)
        self.boxlayout.addWidget(self.tableView)
        # self.setCentralWidget(self.splitter)
        self.setLayout(self.boxlayout)

        self.spinner.setRange(1, pages)
        # self.spinner.setValue(1)
        self.spinner.valueChanged.connect(self.onValueChanged)
        self.count_from_zero.stateChanged.connect(self.onStateCountFromZero)

        self.dataset.signal_pick.connect(self.on_row_pick)
        if self.dataset.get_current_row() is not None:
            self.on_row_pick(self.dataset, self.dataset.get_current_row())
        self._check_pages()

    def onStateCountFromZero(self, state):
        self.tableModel.row_count_start = 0 if state == QtCore.Qt.Checked else 1
        self._update_info()
        self.update()
        # hide/show to force an update
        self.tableView.hide()
        self.tableView.show()
        # self.tableView.repaint()
        # self.tableModel.update()
        # self.resize(self.size())

    def _check_pages(self):
        pages = int(math.ceil(len(self.dataset) * 1. / self.tableModel.page_size))
        current_page = self.spinner.value() - 1
        # if 2 pages, pages 0 and 1 are ok..
        if current_page > pages - 1:  # oops, spinner has invalid value
            self.spinner.setValue(pages + 1)  # then again, gui has page+1
        self._update_info()

    def _update_info(self):
        pages = int(math.ceil(len(self.dataset) * 1. / self.tableModel.page_size))
        current_page = self.spinner.value() - 1
        row_from = current_page * self.tableModel.page_size + self.tableModel.row_count_start
        row_to = row_from + self.tableModel.page_size - 1
        if row_to > len(self.dataset):
            row_to = len(self.dataset) - 1 + self.tableModel.row_count_start
        info_text = "of {pages:,} (row {row_from:,} to {row_to:,})".format(**locals())
        self.label_postfix.setText(info_text)
        self.spinner.setRange(1, pages)

        rows_visible = len(self.dataset) if pages == 1 else self.tableModel.page_size

        info_text_pre = "showing {rows_visible:,} rows on page".format(**locals())
        self.label_prefix.setText(info_text_pre)

    def on_row_pick(self, dataset, row):
        self._check_pages()
        if row is None:
            return
        page = int(row * 1. / self.tableModel.page_size)
        row_offset = row - page * self.tableModel.page_size
        # selection_model = self.tableView.selectionModel()
        # item_model = QtGui.QAbstractItemModel()
        # index = item_model.createIndex(row_offset)
        # selection_model.setCurrentIndex(index)
        self.tableView.selectRow(row_offset)
        self.spinner.setValue(page + 1)  # in GUI, pages starts from 1
        self._update_info()
        # self.tableView.resizeColumnToContents(0)

    def onValueChanged(self, value):
        self.tableModel.set_page(value - 1)  # in GUI, pages start counting from 1, in the code from 0
        self._check_row_select()
        self._update_info()
        # next two lines make the first column (row index) fit to with under osx (bug?)
        # self.tableView.verticalHeader().resizeSections(QtGui.QHeaderView.Stretch)
        # self.tableView.verticalHeader().resizeSections(QtGui.QHeaderView.ResizeToContents)
        self.update()

    def _check_row_select(self):
        row_index = self.dataset.selected_row_index
        if row_index is None:
            return
        page = int(row_index / self.tableModel.page_size)
        page_current = self.spinner.value() - 1
        page_index = row_index - page * self.tableModel.page_size
        logger.debug("row_index=%d page=%d page_index=%d page_current=%d" % (row_index, page, page_index, page_current))
        if page == page_current:  # we are in the right page
            logger.debug("select")
            self.tableView.selectRow(page_index)
        else:
            logger.debug("deselect")
            # self.tableView.selectRow(0) # deselect
            self.tableView.selectionModel().clear()
        self.update()

    def onSelectRow(self, model):
        row_index = model.row() + self.tableModel.get_row_offset()
        logger.debug("row index selected %d" % row_index)
        self.dataset.set_current_row(row_index)



# -*- coding: utf-8 -*-
from __future__ import print_function
from string import Template
import logging
from vaex.ui.qt import *
import numpy as np
import vaex.ui.colormaps
import time

from OpenGL.GL import *  # import GL
from OpenGL.GL.framebufferobjects import *
from OpenGL.GL.ARB.shadow import *
from OpenGL.GL import shaders
import OpenGL

try:
    from PyQt4 import QtGui, QtCore
    from PyQt4 import QtOpenGL
except ImportError:
    try:
        from PyQt5 import QtGui, QtCore
        from PyQt5 import QtOpenGL
    except ImportError:
        from PySide import QtGui, QtCore
        from PySide import QtOpenGL

# import vaex.dataset

logger = logging.getLogger("vaex.ui.volr")


ray_cast_fragment_source = Template("""
#version 120
varying vec4 vertex_color;
uniform sampler1D texture_colormap;
uniform sampler2D texture;
uniform sampler3D cube;
uniform sampler3D gradient;
uniform vec2 size; // size of screen/fbo, to convert between pixels and uniform
//uniform vec2 minmax2d;
uniform vec2 minmax3d;
//uniform vec2 minmax3d_total;
//uniform float maxvalue2d;
//uniform float maxvalue3d;
uniform float alpha_mod; // mod3
uniform float mod4;  // mafnifier
uniform float mod5; // blend color and line integral
uniform float mod6;

uniform sampler1D transfer_function;

uniform float brightness;
uniform float ambient_coefficient;
uniform float diffuse_coefficient;
uniform float specular_coefficient;
uniform float specular_exponent;
uniform float background_opacity;
uniform float foreground_opacity;
uniform float depth_peel;


void main() {
    int steps = ${iterations};
    vec3 ray_end = vec3(texture2D(texture, vec2(gl_FragCoord.x/size.x, gl_FragCoord.y/size.y)));
    vec3 ray_start = vertex_color.xyz;
    //ray_start.z = 1. - ray_start.z;
    //ray_end.z = 1. - ray_end.z;
    float length = 0.;
    vec3 ray_dir = (ray_end - ray_start);
    vec3 ray_delta = ray_dir / float(steps);
    float ray_length = sqrt(ray_dir.x*ray_dir.x + ray_dir.y*ray_dir.y + ray_dir.z*ray_dir.z);
    vec3 ray_pos = ray_start;
    float value = 0.;
    //mat3 direction_matrix = inverse(mat3(transpose(inverse(gl_ModelViewProjectionMatrix))));
    mat3 mat_temp = mat3(gl_ModelViewProjectionMatrix[0].xyz, gl_ModelViewProjectionMatrix[1].xyz, gl_ModelViewProjectionMatrix[2].xyz);
    mat3 direction_matrix = mat_temp;
    vec3 light_pos = (vec3(-100.,100., -100) * direction_matrix).zyx;
    //vec3 light_pos = (direction_matrix * vec3(-5.,5., -100));
    //vec3 origin = (direction_matrix * vec3(0., 0., 0)).xyz;
    vec3 origin = (vec4(0., 0., 0., 0.)).xyz;
    //vec3 light_pos = (vec4(-1000., 0., -1000, 1.)).xyz;
    //mat3 mod = inverse(mat3(gl_ModelViewProjectionMatrix));
    vec4 color = vec4(0, 0, 0, 0);
    vec3 light_dir = light_pos - origin;
    mat3 rotation = mat3(gl_ModelViewMatrix);
    light_dir = vec3(-1,-1,1) * rotation;
    light_dir = normalize(light_dir);// / sqrt(light_dir.x*light_dir.x + light_dir.y*light_dir.y + light_dir.z*light_dir.z);
    float alpha_total = 0.;
    //float normalize = log(maxvalue);
    float intensity_total;
    float data_min = minmax3d.x;
    float data_max = minmax3d.y;
    float data_scale = 1./(data_max - data_min);
    float delta = 1.0/256./2;
    //vec3 light_dir = vec3(1,1,-1);
    vec3 eye = vec3(0, 0, 1) * rotation;
    float depth_factor = 0.;

    for(int i = 0; i < ${iterations}; i++) {
        /*vec3 normal = texture3D(gradient, ray_pos).zyx;
        normal = normal/ sqrt(normal.x*normal.x + normal.y*normal.y + normal.z*normal.z);
        float cosangle = dot(light_dir, normal);
        cosangle = clamp(cosangle, 0., 1.);*/


        vec4 sample = texture3D(cube, ray_pos);
        vec4 sample_x = texture3D(cube, ray_pos + vec3(delta, 0, 0));
        vec4 sample_y = texture3D(cube, ray_pos + vec3(0, delta, 0));
        vec4 sample_z = texture3D(cube, ray_pos + vec3(0, 0, delta));
        if(1==1) {
            vec3 normal = normalize(vec3((sample_x[0]-sample[0])/delta, (sample_y[0]-sample[0])/delta, (sample_z[0]-sample[0])/delta));
            normal = -vec3(normal.x, normal.y, -normal.z);
            //normal *= -sign(normal.z);
            // since the 'surfaces' have two sides, the absolute of the normal is fine
            float cosangle_light = max(abs(dot(light_dir, normal)), 0.);
            float cosangle_eye = max(abs(dot(eye, normal)), 0.);

            float data_value = (sample[0]/1000. - data_min) * data_scale;
            depth_factor = clamp(depth_factor + (data_value > depth_peel ? 1. : 0.), 0., 1.);
            vec4 color_sample = texture1D(transfer_function, data_value);


            //vec4 color_sample = texture1D(texture_colormap, data_value);// * clamp(cosangle, 0.1, 1.);
            float alpha_sample = color_sample.a * sign(data_value) * sign(1.-data_value) / float(steps) * 100.* ray_length; //function_opacities[j]*intensity * sign(data_value) * sign(1.-data_value) / float(steps) * 100.* ray_length ;//clamp(1.-chisq, 0., 1.) * 0.5;//1./128.* length(color_sample) * 100.;
            alpha_sample = clamp(alpha_sample * foreground_opacity, 0., 1.);
            alpha_sample *= depth_factor;
            color_sample = color_sample * (ambient_coefficient + diffuse_coefficient*cosangle_light + specular_coefficient * pow(cosangle_eye, specular_exponent));
            //color_sample = vec4(normal, 1.);
            color = color + (1.0 - alpha_total) * color_sample * alpha_sample;
            alpha_total = clamp(alpha_total + alpha_sample, 0., 1.);
            if(alpha_total >= 1.)
                break;
        }
        if(1==1) {
            vec3 normal = normalize(-vec3((sample_x[1]-sample[1])/delta, (sample_y[1]-sample[1])/delta, (sample_z[1]-sample[1])/delta));
            normal = -vec3(normal.x, normal.y, -normal.z);
            float cosangle_light = max(dot(light_dir, normal), 0.);
            float cosangle_eye = max(dot(eye, normal), 0.);

            float data_value = (sample[1]/1000. - data_min) * data_scale;
            vec4 color_sample = texture1D(transfer_function, data_value);

            //vec4 color_sample = texture1D(texture_colormap, data_value);// * clamp(cosangle, 0.1, 1.);
            float alpha_sample = color_sample.a * sign(data_value) * sign(1.-data_value) / float(steps) * 100.* ray_length; //function_opacities[j]*intensity * sign(data_value) * sign(1.-data_value) / float(steps) * 100.* ray_length ;//clamp(1.-chisq, 0., 1.) * 0.5;//1./128.* length(color_sample) * 100.;
            alpha_sample = clamp(alpha_sample * background_opacity, 0., 1.);
            color_sample = color_sample * (ambient_coefficient + diffuse_coefficient*cosangle_light + specular_coefficient * pow(cosangle_eye, specular_exponent));
            //color_sample = vec4(normal, 1.);
            color = color + (1.0 - alpha_total) * color_sample * alpha_sample;
            alpha_total = clamp(alpha_total + alpha_sample, 0., 1.);
            if(alpha_total >= 1.)
                break;
        }
        ray_pos += ray_delta;
    }
    gl_FragColor = vec4(color.rgb, alpha_total) * brightness; //brightness;
    //gl_FragColor = vec4(ray_pos.xyz, 1) * 100.; //brightness;
}
""")


# print GL_R32F
GL_R32F = 33326


class VolumeRenderWidget(QtOpenGL.QGLWidget):
    def __init__(self, parent=None, function_count=3):
        super(VolumeRenderWidget, self).__init__(parent)
        self.mouse_button_down = False
        self.mouse_button_down_right = False
        self.mouse_x, self.mouse_y = 0, 0
        self.angle1 = 0
        self.angle2 = 0
        self.mod1 = 0
        self.mod2 = 0
        self.mod3 = 0
        self.mod4 = 0
        self.mod5 = 0
        self.mod6 = 0

        self.orbit_angle = 0
        self.orbit_delay = 50
        self.orbiting = False

        self.function_count = function_count
        self.function_opacities = [0.1 / 2**(function_count - 1 - k) for k in range(function_count)]
        self.function_sigmas = [0.05] * function_count
        self.function_means = (np.arange(function_count) / float(function_count - 1)) * 0.8 + 0.10

        self.brightness = 2.
        self.min_level = 0.
        self.max_level = 1.

        self.min_level_vector3d = 0.
        self.max_level_vector3d = 1.
        self.vector3d_scale = 1.
        self.vector3d_auto_scale = True
        self.vector3d_auto_scale_scale = 1.
        self.texture_function_size = 1024 * 4

        self.ambient_coefficient = 0.5
        self.diffuse_coefficient = 0.8
        self.specular_coefficient = 0.5
        self.specular_exponent = 5.
        self.draw_vectors = True
        self.background_opacity = 0.1
        self.foreground_opacity = 1.

        self.texture_cube, self.texture_gradient = None, None
        self.setMouseTracking(True)
        shortcut = QtGui.QShortcut(QtGui.QKeySequence("space"), self)
        shortcut.activated.connect(self.toggle)
        self.texture_index = 1
        self.colormap_index = 0
        self.texture_size = 512  # *8
        self.grid_gl = None
        # gets executed after initializeGL, can hook up your loading of data here
        self.post_init = lambda: 1

        self.arrow_model = Arrow(0, 0, 0, 4.)
        self.update_timer = QtCore.QTimer(self)
        self.update_timer.timeout.connect(self.orbit_progress)
        self.update_timer.setInterval(1000 / 50)  # max 50 fps to save cpu/battery?
        self.ray_iterations = 500
        self.depth_peel = 0.0
        # self.orbit_start()
        # self.update_timer.setSingleShot(False)

    def set_iterations(self, iterations):
        self.ray_iterations = iterations
        self.shader_ray_cast = self.create_shader_ray_cast()

    def setResolution(self, size):
        self.texture_size = size
        self.makeCurrent()
        # self.textures = self.texture_backside, self.texture_final = glGenTextures(2)
        # print "textures", self.textures
        for texture in [self.texture_backside, self.texture_final]:
            glBindTexture(GL_TEXTURE_2D, texture)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, self.texture_size, self.texture_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)
            glBindTexture(GL_TEXTURE_2D, 0)

        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        glFramebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.texture_backside, 0)
        glBindRenderbuffer(GL_RENDERBUFFER, self.render_buffer)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, self.texture_size, self.texture_size)
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, self.render_buffer)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def orbit_start(self):
        # self.orbiting = True
        self.orbit_time_previous = time.time()
        self.stutter_last = time.time()
        # self.orbit_angle = 0
        # self.update()
        self.update_timer.start()
        # QtCore.QTimer.singleShot(self.orbit_delay, self.orbit_progress)

    def orbit_stop(self):
        self.orbiting = False
        self.update_timer.stop()

    def orbit_progress(self):
        orbit_time_now = time.time()
        delta_time = orbit_time_now - self.orbit_time_previous
        # if self.orbiting:
        self.orbit_angle += delta_time / 4. * 360
        # self.orbit_angle += 1.
        # else:
        #   self.orbit_angle = 0
        self.updateGL()
        glFinish()
        # if self.orbiting:
        # ms_spend = (time.time() - self.orbit_time_previous) * 1000
        # QtCore.QTimer.singleShot(max(0, self.orbit_delay - ms_spend), self.orbit_progress)
        # QtCore.QTimer.singleShot(50, self.orbit_progress)
        # self.orbit_time_previous = time.time() # orbit_time_now
        fps = 1. / delta_time
        if 1:  # fps < 16:
            stutter_time = time.time()
            print(".", fps, stutter_time - self.stutter_last)
            self.stutter_last = stutter_time
        self.orbit_time_previous = orbit_time_now

    def toggle(self, ignore=None):
        print("toggle")
        self.texture_index += 1
        self.update()

    def create_shader_color(self):
        self.vertex_shader_color = shaders.compileShader("""
            varying vec4 vertex_color;
            void main() {
                gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
                vertex_color =  gl_Vertex /80. + vec4(0.5, 0.5, 0.5, 0.);
                vertex_color.z = 1.-vertex_color.z;
            }""", GL_VERTEX_SHADER)
        self.fragment_shader_color = shaders.compileShader("""
            varying vec4 vertex_color;
            void main() {
                gl_FragColor = vertex_color;
            }""", GL_FRAGMENT_SHADER)
        return shaders.compileProgram(self.vertex_shader_color, self.fragment_shader_color)

    def create_shader_ray_cast(self):
        self.vertex_shader = shaders.compileShader("""
            varying vec4 vertex_color;
            void main() {
                gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
                //vertex_color = gl_Color;
                vertex_color =  gl_Vertex.x > 1.5 ? vec4(1,0,0,0) : vec4(0,1,0,0)  ;// vec4(gl_Color) + vec4(1, 0, 0, 0);
                vertex_color =  gl_Vertex /80. + vec4(0.5, 0.5, 0.5, 0.);
                vertex_color.z = 1.-vertex_color.z;
            }""", GL_VERTEX_SHADER)
        self.fragment_shader = shaders.compileShader(ray_cast_fragment_source.substitute(iterations=self.ray_iterations), GL_FRAGMENT_SHADER)

        return shaders.compileProgram(self.vertex_shader, self.fragment_shader, validate=False)

    def create_shader_vectorfield(self):
        self.vertex_shader_color = shaders.compileShader("""
            #extension GL_ARB_draw_instanced : enable
            varying vec4 vertex_color;
            void main() {
                float x = floor(float(gl_InstanceIDARB)/(8.*8.)) + 0.5;
                float y = mod(floor(float(gl_InstanceIDARB)/8.), 8.) + 0.5;
                float z = mod(float(gl_InstanceIDARB), 8.) + 0.5;
                vec4 pos = (gl_Vertex + vec4(x*80./8., y*80./8., z*80./8., 0));
                gl_Position = gl_ModelViewProjectionMatrix * pos;
                vertex_color =  pos /80. + vec4(0.5, 0.5, 0.5, 0.);
                //vertex_color.z = 1. - vertex_color.z;
            }""", GL_VERTEX_SHADER)
        self.fragment_shader_color = shaders.compileShader("""
            varying vec4 vertex_color;
            void main() {
                gl_FragColor = vertex_color;
            }""", GL_FRAGMENT_SHADER)
        return shaders.compileProgram(self.vertex_shader_color, self.fragment_shader_color)

    def create_shader_vectorfield_color(self):
        self.vertex_shader_color = shaders.compileShader("""
        #version 120
            #extension GL_ARB_draw_instanced : enable
            varying vec4 vertex_color;
            uniform sampler3D vectorfield;
            uniform int grid_size;
            uniform int use_light;
            uniform vec3 light_color;
            uniform vec3 lightdir;
            uniform float count_level_min;
            uniform float count_level_max;
            uniform float vector3d_scale;
            uniform float vector3d_auto_scale_scale;

            void main() {
                float grid_size_f = float(grid_size);
                float x = floor(float(gl_InstanceIDARB)/(grid_size_f*grid_size_f))/grid_size_f;
                float y = mod(floor(float(gl_InstanceIDARB)/grid_size_f), grid_size_f)/grid_size_f;
                float z = mod(float(gl_InstanceIDARB), grid_size_f)/grid_size_f;
                vec3 uniform_center = vec3(x, y, z);
                vec4 sample = texture3D(vectorfield, uniform_center.yzx);
                vec3 velocity = sample.xyz;
                velocity.x *= -1.;
                velocity.y *= -1.;
                velocity.z *= 1.;
                float counts = sample.a;
                float scale = (counts >= count_level_min) && (counts <= count_level_max) ? 1. : 0.0;
                float speed = length(velocity);
                vec3 direction = normalize(velocity) ;// / speed;
                // form two orthogonal vector to define a rotation matrix
                // the rotation around the vector's axis doesn't matter
                vec3 some_axis = normalize(vec3(0., 1., 1.));
                //vec3 some_axis2 = normalize(vec3(1., 0., 1.));
                vec3 axis1 = normalize(cross(direction, some_axis));
                // + (1-length(cross(direction, some_axis)))*cross(direction, some_axis2));
                vec3 axis2 = normalize(cross(direction, axis1));
                mat3 rotation_and_scaling = mat3(axis1, axis2, direction * (speed) * 20. * vector3d_scale * vector3d_auto_scale_scale);
                mat3 rotation_and_scaling_inverse_transpose = mat3(axis1, axis2, direction / (speed) / 20. / vector3d_scale / vector3d_auto_scale_scale);


                vec3 pos = gl_Vertex.xyz;//
                pos.z -= 0.5;
                pos.z = -pos.z;
                pos *= scale;
                pos = rotation_and_scaling * pos;
                vec3 center = (uniform_center - vec3(0.5,0.5,0.5) + 1./grid_size_f/2.) * 80.;
                center.z = -center.z;
                vec4 transformed_pos = vec4(pos + center, 1);
                //transformed_pos.z = 80. - transformed_pos.z;
                vertex_color =  transformed_pos/80. + vec4(0.5, 0.5, 0.5, 1.); //vec4(uniform_center + gl_ModelViewMatrix*pos, 0.);// + vec4(0.5, 0.5, 0.0, 1.);
                vertex_color.z = 1. - vertex_color.z;
                gl_Position = gl_ModelViewProjectionMatrix * transformed_pos;
                if(use_light == 1) {
                    float fraction = 0.5;
                    vec3 normal =  normalize(mat3(gl_ModelViewMatrix) * rotation_and_scaling_inverse_transpose * gl_Normal);
                    //vec3 normal = normalize(gl_NormalMatrix * gl_Normal);
                    //mat3 rotation = mat3(m);
                    vec3 lightdir_t = normalize(lightdir);
                    vertex_color = vec4(light_color * fraction + max(dot(lightdir_t, normal), 0.), 1.);
                    //vertex_color = vec4(normal, 1.0); //vec4(lightdir_t, 1.);
                }

            }""", GL_VERTEX_SHADER)
        self.fragment_shader_color = shaders.compileShader("""
            varying vec4 vertex_color;
            void main() {
                gl_FragColor = vertex_color;
            }""", GL_FRAGMENT_SHADER)
        return shaders.compileProgram(self.vertex_shader_color, self.fragment_shader_color)

    def paintGL(self):
        # if self.grid_gl is not None:
        if 1:
            glMatrixMode(GL_MODELVIEW)

            glLoadIdentity()
            glTranslated(0.0, 0.0, -15.0)
            glRotated(self.orbit_angle, 0.0, 1.0, 0.0)
            glRotated(self.angle1, 1.0, 0.0, 0.0)
            glRotated(self.angle2, 0.0, 1.0, 0.0)
            if 0:
                glClearColor(0.0, 0.0, 0.0, 1.0)
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
                glBegin(GL_TRIANGLES)
                glColor3f(1., 0., 0.)
                glVertex3f(-30, -30, -20)
                glColor3f(0., 2., 0.)
                glVertex3f(30, -30, -20)
                glColor3f(0., 0., 1.)
                glVertex3f(0, 15, -20)
                glEnd()
            else:
                if self.grid_gl is not None:
                    self.draw_backside()
                    self.draw_frontside()
                    self.draw_to_screen()
                else:
                    glViewport(0, 0, self.texture_size, self.texture_size)
                    glBindFramebuffer(GL_FRAMEBUFFER, 0)
                    glClearColor(0.0, 0.0, 0.0, 1.0)
                    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        # glFlush()
        # glFinish()

    def draw_backside(self):
        glViewport(0, 0, self.texture_size, self.texture_size)
        # glViewport(0, 0, 128*2, 128*2)
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.texture_backside, 0)
        # glClearColor(1.0, 1.0, 0.0, 1.0)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glEnable(GL_DEPTH_TEST)

        glEnable(GL_CULL_FACE)
        glCullFace(GL_FRONT)

        glShadeModel(GL_SMOOTH)
        glUseProgram(self.shader_color)
        self.cube(size=80)
        # self.cube(size=80, gl_type=GL_LINE_LOOP)
        self.wireframe(size=80.)
        glCullFace(GL_BACK)
        if 0:
            self.cube(size=10)  # 'debug' cube
        # self.arrow(0.5, 0.5, 0.5, 80.0)
        # self.arrow_model.drawGL()
        glUseProgram(self.shader_vectorfield_color)
        if self.vectorgrid is not None and self.draw_vectors:
            loc = glGetUniformLocation(self.shader_vectorfield_color, "vectorfield")
            glUniform1i(loc, 0)  # texture unit 0
            loc = glGetUniformLocation(self.shader_vectorfield_color, "grid_size")
            glUniform1i(loc, self.vectorgrid.shape[0])
            loc = glGetUniformLocation(self.shader_vectorfield_color, "use_light")
            glUniform1i(loc, 0)
            mi, ma = np.nanmin(self.vectorgrid_counts), np.nanmax(self.vectorgrid_counts)
            # loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_min");
            # glUniform1f(loc, mi + (ma-mi) * self.min_level_vector3d)
            # loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_max");
            # glUniform1f(loc, mi + (ma-mi) * self.max_level_vector3d)

            loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_min")
            glUniform1f(loc, 10**(np.log10(ma) * self.min_level_vector3d))
            loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_max")
            glUniform1f(loc, 10**(np.log10(ma) * self.max_level_vector3d))

            loc = glGetUniformLocation(self.shader_vectorfield_color, "vector3d_scale")
            glUniform1f(loc, self.vector3d_scale)

            loc = glGetUniformLocation(self.shader_vectorfield_color, "vector3d_auto_scale_scale")
            glUniform1f(loc, self.vector3d_auto_scale_scale if self.vector3d_auto_scale else 1.)

            glActiveTexture(GL_TEXTURE0)
            glEnable(GL_TEXTURE_3D)
            glBindTexture(GL_TEXTURE_3D, self.texture_cube_vector)
            self.arrow_model.drawGL(self.vectorgrid.shape[0]**3)
        glUseProgram(0)
        glActiveTexture(GL_TEXTURE0)
        glDisable(GL_TEXTURE_3D)
        # return

    def arrow(self, x, y, z, scale):
        headfraction = 0.4
        baseradius = 0.1 * scale
        headradius = 0.2 * scale

        # draw base
        glBegin(GL_QUADS)
        # glColor3f(1., 0., 0.)
        for part in range(10):
            angle = np.radians(part / 10. * 360)
            angle2 = np.radians((part + 1) / 10. * 360)
            glNormal3f(np.cos(angle), np.sin(angle), 0.)
            glVertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z + scale / 2 - headfraction * scale)
            glNormal3f(np.cos(angle), np.sin(angle), 0.)
            glVertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z - scale / 2)
            glNormal3f(np.cos(angle2), np.sin(angle2), 0.)
            glVertex3f(x + baseradius * np.cos(angle2), y + baseradius * np.sin(angle2), z - scale / 2)
            glNormal3f(np.cos(angle2), np.sin(angle2), 0.)
            glVertex3f(x + baseradius * np.cos(angle2), y + baseradius * np.sin(angle2), z + scale / 2 - headfraction * scale)
        glEnd()
        glBegin(GL_TRIANGLE_FAN)
        glNormal3f(0, 0, -1)
        # glColor3f(0., 1., 0.)
        glVertex3f(x, y, z - scale / 2)
        for part in range(10 + 1):
            angle = np.radians(-part / 10. * 360)
            glVertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z - scale / 2)
        glEnd()

        glBegin(GL_TRIANGLES)
        # glColor3f(0., 0., 1.)
        a = headradius - baseradius
        b = headfraction * scale
        headangle = np.arctan(a / b)
        for part in range(10 + 1):
            angle = np.radians(-part / 10. * 360)
            anglemid = np.radians(-(part + 0.5) / 10. * 360)
            angle2 = np.radians(-(part + 1) / 10. * 360)
            glNormal3f(np.cos(anglemid) * np.cos(headangle), np.sin(anglemid) * np.cos(headangle), np.sin(headangle))
            glVertex3f(x, y, z + scale / 2)

            glNormal3f(np.cos(angle2) * np.cos(headangle), np.sin(angle2) * np.cos(headangle), np.sin(headangle))
            glVertex3f(x + headradius * np.cos(angle2), y + headradius * np.sin(angle2), z + scale / 2 - headfraction * scale)

            glNormal3f(np.cos(angle) * np.cos(headangle), np.sin(angle) * np.cos(headangle), np.sin(headangle))
            glVertex3f(x + headradius * np.cos(angle), y + headradius * np.sin(angle), z + scale / 2 - headfraction * scale)
        glEnd()

        glBegin(GL_QUADS)
        # glColor3f(1., 1., 0.)
        glNormal3f(0, 0, -1)
        for part in range(10):
            angle = np.radians(part / 10. * 360)
            angle2 = np.radians((part + 1) / 10. * 360)
            glVertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z + scale / 2 - headfraction * scale)
            glVertex3f(x + baseradius * np.cos(angle2), y + baseradius * np.sin(angle2), z + scale / 2 - headfraction * scale)
            glVertex3f(x + headradius * np.cos(angle2), y + headradius * np.sin(angle2), z + scale / 2 - headfraction * scale)
            glVertex3f(x + headradius * np.cos(angle), y + headradius * np.sin(angle), z + scale / 2 - headfraction * scale)
        glEnd()

    def draw_frontside(self):
        glViewport(0, 0, self.texture_size, self.texture_size)
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        # glBindFramebuffer(GL_FRAMEBUFFER, 0)

        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.texture_final, 0)

        glClearColor(1.0, 1.0, 1.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glEnable(GL_CULL_FACE)
        glCullFace(GL_BACK)

        glShadeModel(GL_SMOOTH)

        glDisable(GL_BLEND)
        glColor3f(0, 0, 0)
        # self.cube(size=80, gl_type=GL_LINE_LOOP, color_axis=True) # draw the wireframe cube
        self.wireframe(size=80.)
        if 0:
            self.cube(size=10)  # 'debug' cube
        glEnable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        g = 0.5
        glMaterialfv(GL_FRONT, GL_SPECULAR, [g, g, g, 1.])
        glMaterialfv(GL_FRONT, GL_DIFFUSE, [g, g, g, 1.])
        glPushMatrix()
        glLoadIdentity()
        # glLightfv(GL_LIGHT0, GL_POSITION, [1, -0.5, 1, 0.])
        glLightfv(GL_LIGHT0, GL_POSITION, [0.1, 0.1, 1, 0.])
        glPopMatrix()
        a = 0.5
        glLightfv(GL_LIGHT0, GL_AMBIENT, [a, a, a, 0.])
        glMaterialfv(GL_FRONT, GL_SHININESS, [50.])
        glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE)
        glEnable(GL_COLOR_MATERIAL)
        glColor3f(0.5, 0, 0)
        # self.arrow(0.5, 0.5, 0.5, 80.0)
        if self.vectorgrid is not None and self.draw_vectors:
            glUseProgram(self.shader_vectorfield_color)
            loc = glGetUniformLocation(self.shader_vectorfield_color, "vectorfield")
            glUniform1i(loc, 0)  # texture unit 0
            loc = glGetUniformLocation(self.shader_vectorfield_color, "grid_size")
            glUniform1i(loc, self.vectorgrid.shape[0])
            loc = glGetUniformLocation(self.shader_vectorfield_color, "use_light")
            glUniform1i(loc, 1)
            loc = glGetUniformLocation(self.shader_vectorfield_color, "light_color")
            glUniform3f(loc, 1., 0., 0.)
            loc = glGetUniformLocation(self.shader_vectorfield_color, "lightdir")
            glUniform3f(loc, -1., -1., 1.)
            mi, ma = np.nanmin(self.vectorgrid_counts), np.nanmax(self.vectorgrid_counts)
            # loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_min");
            # //glUniform1f(loc, mi + (ma-mi) * self.min_level_vector3d)
            # glUniform1f(loc, 10**(log10(ma)*self.min_level_vector3d)
            # loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_max");
            # glUniform1f(loc, mi + (ma-mi) * self.max_level_vector3d)

            loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_min")
            glUniform1f(loc, 10**(np.log10(ma) * self.min_level_vector3d))
            loc = glGetUniformLocation(self.shader_vectorfield_color, "count_level_max")
            glUniform1f(loc, 10**(np.log10(ma) * self.max_level_vector3d))

            loc = glGetUniformLocation(self.shader_vectorfield_color, "vector3d_scale")
            glUniform1f(loc, self.vector3d_scale)

            loc = glGetUniformLocation(self.shader_vectorfield_color, "vector3d_auto_scale_scale")
            glUniform1f(loc, self.vector3d_auto_scale_scale if self.vector3d_auto_scale else 1.)

            glActiveTexture(GL_TEXTURE0)
            glEnable(GL_TEXTURE_3D)
            glBindTexture(GL_TEXTURE_3D, self.texture_cube_vector)
            self.arrow_model.drawGL(self.vectorgrid.shape[0]**3)
            glDisable(GL_TEXTURE_3D)
        glUseProgram(0)
        glDisable(GL_LIGHTING)
        glDisable(GL_LIGHT0)
        glEnable(GL_BLEND)

        glBlendEquation(GL_FUNC_ADD, GL_FUNC_ADD)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glUseProgram(self.shader_ray_cast)
        loc = glGetUniformLocation(self.shader_ray_cast, "texture")
        glUniform1i(loc, 0)  # texture unit 0
        glBindTexture(GL_TEXTURE_2D, self.texture_backside)
        glEnable(GL_TEXTURE_2D)
        glActiveTexture(GL_TEXTURE0)

        loc = glGetUniformLocation(self.shader_ray_cast, "cube")
        glUniform1i(loc, 1)  # texture unit 1
        glActiveTexture(GL_TEXTURE1)
        glBindTexture(GL_TEXTURE_3D, self.texture_cube)
        # glEnable(GL_TEXTURE_3D)

        loc = glGetUniformLocation(self.shader_ray_cast, "texture_colormap")
        glUniform1i(loc, 2)  # texture unit 2
        glActiveTexture(GL_TEXTURE2)
        # index = vaex.ui.colormaps.colormaps.index("afmhot")
        index = 16
        glBindTexture(GL_TEXTURE_1D, self.textures_colormap[index])
        glEnable(GL_TEXTURE_1D)

        if 1:
            loc = glGetUniformLocation(self.shader_ray_cast, "transfer_function")
            glUniform1i(loc, 3)  # texture unit 3
            glActiveTexture(GL_TEXTURE3)
            # index = vaex.ui.colormaps.colormaps.index("afmhot")
            glBindTexture(GL_TEXTURE_1D, self.texture_function)
            rgb = self.colormap_data[self.colormap_index]
            x = np.arange(self.texture_function_size) / (self.texture_function_size - 1.)
            y = x * 0.
            for i in range(3):
                y += np.exp(-((x - self.function_means[i]) / self.function_sigmas[i])**2) * self.function_opacities[i]
                # y +=np.exp(-((nx-self.function_means[i])/self.function_sigmas[i])**2) * (np.log10(self.function_opacities[i])+3)/3 * 32.
            # print "max opacity", np.max(y)
            self.function_data[:, 0] = rgb[:, 0]
            self.function_data[:, 1] = rgb[:, 1]
            self.function_data[:, 2] = rgb[:, 2]
            self.function_data[:, 3] = (y * 255).astype(np.uint8)
            self.function_data_1d = self.function_data.reshape(-1)
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA8, self.texture_function_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, self.function_data_1d)

            glEnable(GL_TEXTURE_1D)

        if 0:
            loc = glGetUniformLocation(self.shader_ray_cast, "gradient")
            glUniform1i(loc, 4)  # texture unit 4
            glActiveTexture(GL_TEXTURE4)
            glEnable(GL_TEXTURE_3D)
            glBindTexture(GL_TEXTURE_3D, self.texture_gradient)

        glActiveTexture(GL_TEXTURE0)

        size = glGetUniformLocation(self.shader_ray_cast, "size")
        glUniform2f(size, self.texture_size, self.texture_size)

        depth_peel = glGetUniformLocation(self.shader_ray_cast, "depth_peel")
        glUniform1f(depth_peel, self.depth_peel)

        # maxvalue = glGetUniformLocation(self.shader_ray_cast,"maxvalue");
        # glUniform1f(maxvalue, self.data3d.max()*10**self.mod2);

        # minmax = glGetUniformLocation(self.shader_ray_cast,"minmax2d");
        # glUniform2f(minmax, 1*10**self.mod1, self.grid2d_max*10**self.mod2);

        minmax = glGetUniformLocation(self.shader_ray_cast, "minmax3d")
        # xmin =  self.grid_min +  (self.grid_max -  self.grid_min) * self.min_level;
        # xmax =  self.grid_min +  (self.grid_max -  self.grid_min) * self.max_level;
        glUniform2f(minmax, self.min_level, self.max_level)

        # minmax3d_total = glGetUniformLocation(self.shader_ray_cast,"minmax3d_total");
        # glUniform2f(minmax3d_total, self.grid_min, self.grid_max);

        glUniform1f(glGetUniformLocation(self.shader_ray_cast, "brightness"), self.brightness)
        glUniform1f(glGetUniformLocation(self.shader_ray_cast, "background_opacity"), self.background_opacity)
        glUniform1f(glGetUniformLocation(self.shader_ray_cast, "foreground_opacity"), self.foreground_opacity)

        glUniform1fv(glGetUniformLocation(self.shader_ray_cast, "function_means"), self.function_count, self.function_means)
        glUniform1fv(glGetUniformLocation(self.shader_ray_cast, "function_sigmas"), self.function_count, self.function_sigmas)
        glUniform1fv(glGetUniformLocation(self.shader_ray_cast, "function_opacities"), self.function_count, self.function_opacities)

        for name in ["ambient_coefficient", "diffuse_coefficient", "specular_coefficient", "specular_exponent"]:
            glUniform1f(glGetUniformLocation(self.shader_ray_cast, name), getattr(self, name))

        alpha_mod = glGetUniformLocation(self.shader_ray_cast, "alpha_mod")
        glUniform1f(alpha_mod, 10**self.mod3)

        for i in range(4, 7):
            name = "mod" + str(i)
            mod = glGetUniformLocation(self.shader_ray_cast, name)
            glUniform1f(mod, 10**getattr(self, name))

        self.shader_ray_cast.check_validate()

        glShadeModel(GL_SMOOTH)
        self.cube(size=80)  # do the volume rendering
        glUseProgram(0)

        glActiveTexture(GL_TEXTURE4)
        glBindTexture(GL_TEXTURE_3D, 0)
        glEnable(GL_TEXTURE_2D)

        glActiveTexture(GL_TEXTURE3)
        glBindTexture(GL_TEXTURE_1D, 0)
        glEnable(GL_TEXTURE_2D)

        glActiveTexture(GL_TEXTURE2)
        glBindTexture(GL_TEXTURE_1D, 0)
        glEnable(GL_TEXTURE_2D)

        glActiveTexture(GL_TEXTURE1)
        glBindTexture(GL_TEXTURE_2D, 0)
        glEnable(GL_TEXTURE_2D)

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, 0)
        glEnable(GL_TEXTURE_2D)

        glDisable(GL_BLEND)
        # return

    def draw_to_screen(self):
        w = self.width()
        h = self.height()
        glViewport(0, 0, w, h)
        # glShadeModel(GL_FLAT);

        glBindFramebuffer(GL_FRAMEBUFFER, 0)
        glClearColor(1.0, 0.0, 0.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glCullFace(GL_BACK)

        glBindTexture(GL_TEXTURE_2D, self.textures[self.texture_index % len(self.textures)])
        # glBindTexture(GL_TEXTURE_3D, self.texture_cube)
        glEnable(GL_TEXTURE_2D)
        glLoadIdentity()
        glBegin(GL_QUADS)
        w = 50
        z = -1
        glTexCoord2f(0, 0)
        glVertex3f(-w, -w, z)
        glTexCoord2f(1, 0)
        glVertex3f(w, -w, z)
        glTexCoord2f(1, 1)
        glVertex3f(w, w, z)
        glTexCoord2f(0, 1)
        glVertex3f(-w, w, z)
        glEnd()
        glBindTexture(GL_TEXTURE_2D, 0)

        # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    def draw_to_screen_(self):
        w = self.width()
        h = self.height()
        glViewport(0, 0, w, h)
        # glShadeModel(GL_FLAT);

        glBindFramebuffer(GL_FRAMEBUFFER, 0)
        glClearColor(0.0, 1.0, 0.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # glCullFace(GL_BACK);

        # glBindTexture(GL_TEXTURE_2D, self.textures[self.texture_index % len(self.textures)])
        glBindTexture(GL_TEXTURE_1D, 0)
        glBindTexture(GL_TEXTURE_2D, 0)
        glEnable(GL_TEXTURE_3D)
        glBindTexture(GL_TEXTURE_3D, self.texture_cube)
        glEnable(GL_TEXTURE_3D)
        glColor3f(1, 0, 0)
        glLoadIdentity()
        glBegin(GL_QUADS)
        w = 20
        z = -1
        glTexCoord3f(0, 0, 0.5)
        glVertex3f(-w, -w, z)
        glTexCoord3f(1, 0, 0.5)
        glVertex3f(w, -w, z)
        glTexCoord3f(1, 1, 0.5)
        glVertex3f(w, w, z)
        glTexCoord3f(0, 1, 0.5)
        glVertex3f(-w, w, z)
        glEnd()
        glBindTexture(GL_TEXTURE_3D, 0)

        # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    def wireframe(self, size, color_axis=False):
        w = size / 2
        glLineWidth(2.)
        # X
        glBegin(GL_LINES)
        glColor3f(1., 0., 0.)
        glVertex3f(-w, -w, w)
        glVertex3f(w, -w, w)
        glEnd()

        # Y
        glBegin(GL_LINES)
        glColor3f(0., 1., 0.)
        glVertex3f(-w, -w, w)
        glVertex3f(-w, w, w)
        glEnd()

        # Z
        glBegin(GL_LINES)
        glColor3f(0., 0., 1.)
        glVertex3f(-w, -w, -w)
        glVertex3f(-w, -w, w)
        glEnd()

        # the rest thin black
        glLineWidth(1.)
        glColor3f(0., 0., 0.)

        # z parallel
        glBegin(GL_LINES)
        glVertex3f(-w, w, -w)
        glVertex3f(-w, w, w)
        glEnd()
        glBegin(GL_LINES)
        glVertex3f(w, w, -w)
        glVertex3f(w, w, w)
        glEnd()
        glBegin(GL_LINES)
        glVertex3f(w, -w, -w)
        glVertex3f(w, -w, w)
        glEnd()

        # Y parallel
        glBegin(GL_LINES)
        glVertex3f(w, -w, -w)
        glVertex3f(w, w, -w)
        glEnd()
        glBegin(GL_LINES)
        glVertex3f(w, -w, w)
        glVertex3f(w, w, w)
        glEnd()
        glBegin(GL_LINES)
        glVertex3f(-w, -w, -w)
        glVertex3f(-w, w, -w)
        glEnd()

        # X parallel
        glBegin(GL_LINES)

        glVertex3f(-w, w, -w)
        glVertex3f(w, w, -w)
        glEnd()
        glBegin(GL_LINES)
        glVertex3f(-w, -w, -w)
        glVertex3f(w, -w, -w)
        glEnd()
        glBegin(GL_LINES)
        glVertex3f(-w, w, w)
        glVertex3f(w, w, w)
        glEnd()

    def cube(self, size, gl_type=GL_QUADS):
        w = size / 2.

        def vertex(x, y, z):
            # glColor3f(x+w, y+w, z+w)
            # glMultiTexCoord3f(GL_TEXTURE1, x, y, z);
            glVertex3f(x, y, z)

        # back
        if 1:
            # glColor3f(1, 0, 0)
            glBegin(gl_type)
            vertex(-w, -w, -w)
            vertex(-w, w, -w)
            vertex(w, w, -w)
            vertex(w, -w, -w)
            glEnd()

        # front
        if 1:
            glBegin(gl_type)
            # glColor3f(0, 1, 0)
            vertex(-w, -w, w)
            vertex(w, -w, w)
            vertex(w, w, w)
            vertex(-w, w, w)
            glEnd()

        # right
        if 1:
            glBegin(gl_type)
            # glColor3f(0, 0, 1)
            vertex(w, -w, w)
            vertex(w, -w, -w)
            vertex(w, w, -w)
            vertex(w, w, w)
            glEnd()

        # left
        if 1:
            glBegin(gl_type)
            # glColor3f(0, 0, 1)
            vertex(-w, -w, -w)
            vertex(-w, -w, w)
            vertex(-w, w, w)
            vertex(-w, w, -w)
            glEnd()

        # top
        if 1:
            glBegin(gl_type)
            # glColor3f(0, 0, 1)
            vertex(w, w, -w)
            vertex(-w, w, -w)
            vertex(-w, w, w)
            vertex(w, w, w)
            glEnd()

        # bottom
        if 1:
            glBegin(gl_type)
            # glColor3f(0, 0, 1)
            vertex(-w, -w, -w)
            vertex(w, -w, -w)
            vertex(w, -w, w)
            vertex(-w, -w, w)
            glEnd()

    def resizeGL(self, w, h):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        # glOrtho(-50, 50, -50, 50, -50.0, 50.0)
        glOrtho(-50, 50, -50, 50, -150.0, 150.0)
        glViewport(0, 0, w, h)

    def initializeGL(self):

        colormaps = vaex.ui.colormaps.colormaps
        Nx, Ny = self.texture_function_size, 16
        self.colormap_data = np.zeros((len(colormaps), Nx, 3), dtype=np.uint8)

        import matplotlib.cm
        self.textures_colormap = glGenTextures(len(colormaps))
        for i, colormap_name in enumerate(colormaps):
            colormap = matplotlib.cm.get_cmap(colormap_name)
            mapping = matplotlib.cm.ScalarMappable(cmap=colormap)
            # pixmap = QtGui.QPixmap(32*2, 32)
            x = np.arange(Nx) / (Nx - 1.)
            # x = np.vstack([x]*Ny)
            rgba = mapping.to_rgba(x, bytes=True).reshape(Nx, 4)
            rgb = rgba[:, 0:3] * 1
            self.colormap_data[i] = rgb  # (rgb*255.).astype(np.uint8)
            if i == 0:
                print(rgb[0], rgb[-1], end=' ')

            texture = self.textures_colormap[i]
            glBindTexture(GL_TEXTURE_1D, texture)
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB8, Nx, 0, GL_RGB, GL_UNSIGNED_BYTE, self.colormap_data[i])
            # glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, self.texture_size, self.texture_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, None);
            glBindTexture(GL_TEXTURE_1D, 0)

        if 1:
            self.texture_function = glGenTextures(1)
            texture = self.texture_function
            glBindTexture(GL_TEXTURE_1D, texture)
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            self.function_data = np.zeros((self.texture_function_size, 4), dtype=np.uint8)
            x = np.arange(self.texture_function_size) * 255 / (self.texture_function_size - 1.)
            self.function_data[:, 0] = x
            self.function_data[:, 1] = x
            self.function_data[:, 2] = 0
            self.function_data[:, 3] = x
            self.function_data_1d = self.function_data.reshape(-1)
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA8, Nx, 0, GL_RGBA, GL_UNSIGNED_BYTE, self.function_data_1d)
            # glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, self.texture_size, self.texture_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, None);
            glBindTexture(GL_TEXTURE_1D, 0)

        if 1:
            N = 1024 * 4
            self.surface_data = np.zeros((N, 3), dtype=np.uint8)
            self.texture_surface = glGenTextures(1)
            glBindTexture(GL_TEXTURE_1D, self.texture_surface)
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB8, Nx, 0, GL_RGB, GL_UNSIGNED_BYTE, self.surface_data)
            # glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, self.texture_size, self.texture_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, None);
            glBindTexture(GL_TEXTURE_1D, 0)

        if 0:

            f = glCreateShaderObject(GL_FRAGMENT_SHADER)
            fragment_source = "void main(){ gl_FragColor=gl_FragCoord/512.0; }"
            glShaderSource(f, 1, fs, None)
            glCompileShaderARB(f)

            self.program = glCreateProgramObjectARB()
            glAttachObjectARB(self.program, f)

        # glClearColor(0.0, 0.0, 0.0, 1.0)
        # glClear(GL_COLOR_BUFFER_BIT)

        # print(bool(glGenFramebuffers))
        self.fbo = glGenFramebuffers(1)
        # print(self.fbo)
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)

        self.textures = self.texture_backside, self.texture_final = glGenTextures(2)
        print("textures", self.textures)
        for texture in [self.texture_backside, self.texture_final]:
            glBindTexture(GL_TEXTURE_2D, texture)
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            # glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            # glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            # glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, self.texture_size, self.texture_size, 0, GL_RGBA, GL_FLOAT, None);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, self.texture_size, self.texture_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)
            glBindTexture(GL_TEXTURE_2D, 0)

        glFramebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.texture_backside, 0)

        self.render_buffer = glGenRenderbuffers(1)
        glBindRenderbuffer(GL_RENDERBUFFER, self.render_buffer)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, self.texture_size, self.texture_size)

        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, self.render_buffer)
        # from matplotlib import pylab
        # pylab.imshow(np.log((self.data3d.astype(np.float32)).sum(axis=0)+1), cmap='PaulT_plusmin', origin="lower")
        # pylab.show()

        # print(self.texture_size)
        # print(glCheckFramebufferStatus())

        self.shader_ray_cast = self.create_shader_ray_cast()

        self.shader_color = self.create_shader_color()
        self.shader_vectorfield = self.create_shader_vectorfield()
        self.shader_vectorfield_color = self.create_shader_vectorfield_color()
        self.post_init()
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    # only vaex specific code?
    def loadTable(self, args, column_names, grid_size=128, grid_size_vector=16):
        import vaex.vaexfast
        dataset = vaex.dataset.load_file(sys.argv[1])
        x, y, z, vx, vy, vz = [dataset.columns[name] for name in sys.argv[2:]]
        x, y, z, vx, vy, vz = [k.astype(np.float64) - k.mean() for k in [x, y, z, vx, vy, vz]]
        grid3d = np.zeros((grid_size, grid_size, grid_size), dtype=np.float64)
        vectorgrid = np.zeros((4, grid_size_vector, grid_size_vector, grid_size_vector), dtype=np.float64)

        # mi, ma = -30., 30.
        # mi, ma = 45., 55
        # mi, ma = -20, 20
        s = 0.
        mi, ma = -4, 4
        print("histogram3d")
        vaex.vaexfast.histogram3d(x, y, z, None, grid3d, mi + s, ma + s, mi, ma, mi, ma)
        if 0:
            vx = vx - vx.mean()
            vy = vy - vy.mean()
            vz = vz - vz.mean()
        vaex.vaexfast.histogram3d(x, y, z, vx, vectorgrid[0], mi + s, ma + s, mi, ma, mi, ma)
        vaex.vaexfast.histogram3d(x, y, z, vy, vectorgrid[1], mi + s, ma + s, mi, ma, mi, ma)
        vaex.vaexfast.histogram3d(x, y, z, vz, vectorgrid[2], mi + s, ma + s, mi, ma, mi, ma)
        print(vx)
        print(vectorgrid[0])
        print(vaex.vaexfast.resize(vectorgrid[0], 4))
        print(vaex.vaexfast.resize(vectorgrid[1], 4))
        print(vaex.vaexfast.resize(vectorgrid[2], 4))
        print(vaex.vaexfast.resize(grid3d, 4))
        print("$" * 80)
        vectorgrid[3][:] = vaex.vaexfast.resize(grid3d, grid_size_vector)
        for i in range(3):
            vectorgrid[i] /= vectorgrid[3]  # go from weighted to mean

        if 1:
            vmax = max([np.nanmax(vectorgrid[0]), np.nanmax(vectorgrid[1]), np.nanmax(vectorgrid[2])])
            for i in range(3):
                vectorgrid[i] *= 1
        # self.data3d = self.data3d.astype(np.float32)
        # self.data2d = self.data2d.astype(np.float32)
        # self.size3d = 128 # * 4
        # self.data2d = np.zeros((self.size3d, self.size3d)) #.astype(np.float32)
        #
        # self.data2d

        # import scipy.ndimage
        # #self.data3d = 10**scipy.ndimage.gaussian_filter(np.log10(self.data3d+1), 1.5)-1
        # self.data3d = 10**scipy.ndimage.gaussian_filter(np.log10(self.data3d+1), 0.5)-1
        # data3ds = scipy.ndimage.gaussian_filter((self.data3d), 1.5)
        vectorgrid = np.swapaxes(vectorgrid, 0, 3)

        self.setGrid(grid3d, vectorgrid=vectorgrid)

    def setGrid(self, grid, grid_background=None, vectorgrid=None):
        self.mod1 = 0
        self.mod2 = 0
        self.mod3 = 0
        self.mod4 = 0
        self.mod5 = 0
        self.mod6 = 0
        # print "*" * 70
        # print self.grid
        # print "*" * 70
        # self.grid = np.log10(grid.astype(np.float32)+1)
        if vectorgrid is not None:
            self.vectorgrid = vectorgrid.astype(np.float32)
            self.vectorgrid_counts = self.vectorgrid[:, :, :, 3]
        else:
            self.vectorgrid = None

        def normalise(ar):
            mask = ~np.isinf(ar)
            mi, ma = np.nanmin(ar[mask]), np.nanmax(ar[mask])
            res = (ar - mi) / (ma - mi) * 1000.
            res[~mask] = mi
            return res
        if grid_background is not None:
            self.grid_gl = np.zeros(grid.shape + (2,), np.float32)
            self.grid_gl[:, :, :, 0] = normalise(grid.astype(np.float32))
            self.grid_gl[:, :, :, 1] = normalise(grid_background.astype(np.float32))
        else:
            self.grid_gl = np.zeros(grid.shape + (1,), np.float32)
            self.grid_gl[:, :, :, 0] = normalise(grid.astype(np.float32))
        # self.grid_background = grid_background.astype(np.float32)
        # self.grid = np.log10(grid.astype(np.float32)+1)
        # self.grid_min, self.grid_max = np.nanmin(grid), np.nanmax(grid)
        # print "MINMAX" * 10, self.grid_min, self.grid_max
        # grids_2d = [self.grid.sum(axis=i) for i in range(3)]
        # self.grid2d_min, self.grid2d_max = min([np.nanmin(grid) for grid in grids_2d]), max([np.nanmax(grid) for grid in grids_2d])
        # print "3d", self.grid_min, self.grid_max
        # print "2d", self.grid2d_min, self.grid2d_max
        # return
        # data3ds = data3ds.sum(axis=0)
        if 0:
            self.grid_gradient = np.gradient(self.grid)
            length = np.sqrt(self.grid_gradient[0]**2 + self.grid_gradient[1]**2 + self.grid_gradient[2]**2)
            self.grid_gradient[0] = self.grid_gradient[0] / length
            self.grid_gradient[1] = self.grid_gradient[1] / length
            self.grid_gradient[2] = self.grid_gradient[2] / length

            self.grid_gradient_data = np.zeros(self.grid.shape + (3,), dtype=np.float32)
            self.grid_gradient_data[:, :, :, 0] = self.grid_gradient[0]
            self.grid_gradient_data[:, :, :, 1] = self.grid_gradient[1]
            self.grid_gradient_data[:, :, :, 2] = self.grid_gradient[2]
            self.grid_gradient_data[:, :, :, 2] = 1.
            self.grid_gradient = self.grid_gradient_data
            del self.grid_gradient_data
            print(self.grid_gradient.shape)

        # self.data3d -= self.data3d.min()
        # self.data3d /= self.data3d.max()
        # self.data3d = np.log10(self.data3d+1)
        # self.data2d = np.log10(self.data2d+1)

        # self.data3df = (self.data3d * 1.).astype(np.float32)
        # self.data2df = self.data2d * 1.0

        # self.data3d -= self.data3d.min()
        # self.data3d /= self.data3d.max()
        # self.data3d = (self.data3d * 255).astype(np.uint8)

        # self.data2d -= self.data2d.min()
        # self.data2d /= self.data2d.max()
        # self.data2d = (self.data2d * 255).astype(np.uint8)
        # print self.data3d.max()
        for texture in [self.texture_cube, self.texture_gradient]:
            logger.debug("deleting texture: %r", texture)
            if texture is not None:
                glDeleteTextures(int(texture))

        self.texture_cube = glGenTextures(1)
        self.texture_gradient = glGenTextures(1)
        logger.debug("texture: %s %s", self.texture_cube, self.texture_gradient)
        logger.debug("texture types: %s %s", type(self.texture_cube), type(self.texture_gradient))
        # self.texture_square = glGenTextures(1)

        # glActiveTexture(GL_TEXTURE1);
        # glBindTexture(GL_TEXTURE_3D, self.texture_cube)
        # glBindTexture(GL_TEXTURE_3D, self.texture_cube)
        # glBindTexture(GL_TEXTURE_2D, self.texture_square)
        # glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE8, self.size3d, self.size3d, self.size3d, 0,
        # GL_RED, GL_FLOAT, self.data3d)

        if 0:
            self.rgb3d = np.zeros(self.grid.shape + (3,), dtype=np.uint8)
            self.rgb3d[:, :, :, 0] = self.grid
            self.rgb3d[:, :, :, 1] = self.grid
            self.rgb3d[:, :, :, 2] = self.grid

        glBindTexture(GL_TEXTURE_3D, self.texture_cube)
        width, height, depth = grid.shape[::-1]
        print("dims", width, height, depth)
        if grid_background is not None:
            glTexImage3D(GL_TEXTURE_3D, 0, GL_RG32F, width, height, depth, 0,
                         GL_RG, GL_FLOAT, self.grid_gl)
        else:
            glTexImage3D(GL_TEXTURE_3D, 0, GL_R32F, width, height, depth, 0,
                         GL_RED, GL_FLOAT, self.grid_gl)
        # print self.grid, self.texture_cube
        # glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB8, self.size3d, self.size3d, self.size3d, 0,
        # GL_RGB, GL_UNSIGNED_BYTE, self.rgb3d)

        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        if 1:
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER)
        glBindTexture(GL_TEXTURE_3D, 0)

        if self.vectorgrid is not None:
            # print self.vectorgrid
            # print "counts", self.vectorgrid[3]
            assert self.vectorgrid.shape[0] == self.vectorgrid.shape[1] == self.vectorgrid.shape[2], "wrong shape %r" % self.vectorgrid.shape

            # vx, vy, vz = self.vectorgrid[:,:,0], self.vectorgrid[:,:,1], self.vectorgrid[:,:,2]
            # length = numpy.linalg.norm(vx)
            vectorflat = self.vectorgrid.reshape((-1, 4))
            vx, vy, vz, counts = vectorflat.T
            mask = counts > 0
            # length = scipy.stats.nanmedian(np.sqrt(self.vectorgrid[:,:,0][mask]**2 + self.vectorgrid[:,:,1][mask]**2 + self.vectorgrid[:,:,2][mask]**2))
            target_length = np.nanmean(np.sqrt(vx[mask]**2 + vy[mask]**2 + vz[mask]**2))

            self.vector3d_auto_scale_scale = 1. / target_length / self.vectorgrid.shape[0]
            print("#" * 200)
            print(self.vector3d_auto_scale_scale, target_length)
            # dsadsa
            self.texture_cube_vector_size = self.vectorgrid.shape[0]
            self.texture_cube_vector = glGenTextures(1)
            glBindTexture(GL_TEXTURE_3D, self.texture_cube_vector)
            _, width, height, depth = self.vectorgrid.shape[::-1]
            print("dims vector", width, height, depth)
            glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA32F, width, height, depth, 0,
                         GL_RGBA, GL_FLOAT, self.vectorgrid)
            # print self.grid, self.texture_cube
            # glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB8, self.size3d, self.size3d, self.size3d, 0,
            # GL_RGB, GL_UNSIGNED_BYTE, self.rgb3d)

            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE)
            glBindTexture(GL_TEXTURE_3D, 0)

        # gradient
        if 0:
            glBindTexture(GL_TEXTURE_3D, self.texture_gradient)
            glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB32F, width, height, depth, 0,
                         GL_RGB, GL_FLOAT, self.grid_gradient)
            # glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB8, self.size3d, self.size3d, self.size3d, 0,
            # GL_RGB, GL_UNSIGNED_BYTE, self.rgb3d)

            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            if 1:
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER)
            glBindTexture(GL_TEXTURE_3D, 0)

        if 0:
            import pylab
            pylab.subplot(221)
            pylab.imshow(np.log10(grids_2d[0] + 1))
            pylab.subplot(222)
            pylab.imshow(np.log10(grids_2d[1] + 1))
            pylab.subplot(223)
            pylab.imshow(np.log10(grids_2d[2] + 1))
            pylab.subplot(224)
            pylab.imshow(np.log10(self.grid[128] + 1))
            pylab.show()
        self.update()

        # glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE)

    # sys.exit(0)
        # self.back_buffer, self.final_buffer = glGenTextures(2)
        # bo = glGenFramebuffers(1)

    def mouseMoveEvent(self, event):
        x, y = event.x(), event.y()
        dx = x - self.mouse_x
        dy = y - self.mouse_y

        speed = 1.
        speed_mod = 0.1 / 5. / 5.
        if self.mouse_button_down:
            self.angle2 += dx * speed
            self.angle1 += dy * speed
            print(self.angle1, self.angle2)
        if self.mouse_button_down_right:
            if QtGui.QApplication.keyboardModifiers() == QtCore.Qt.NoModifier:
                self.min_level += dx * speed_mod / 10.
                self.max_level += -dy * speed_mod / 10.
                print("mod1/2", self.min_level, self.max_level)
            if (QtGui.QApplication.keyboardModifiers() == QtCore.Qt.AltModifier) or (QtGui.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier):
                self.mod3 += dx * speed_mod
                self.mod4 += -dy * speed_mod
                print("mod3/4", self.mod3, self.mod4)
            if QtGui.QApplication.keyboardModifiers() == QtCore.Qt.ShiftModifier:
                self.mod5 += dx * speed_mod
                self.mod6 += -dy * speed_mod
                print("mod5/6", self.mod5, self.mod6)

        self.mouse_x, self.mouse_y = x, y
        if not self.update_timer.isActive() and (self.mouse_button_down or self.mouse_button_down_right):
            self.updateGL()

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.mouse_button_down = True
        if event.button() == QtCore.Qt.RightButton:
            self.mouse_button_down_right = True

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.mouse_button_down = False
        if event.button() == QtCore.Qt.RightButton:
            self.mouse_button_down_right = False

    def write(self):
        colormap_name = "afmhot"
        import matplotlib.cm
        colormap = matplotlib.cm.get_cmap(colormap_name)
        mapping = matplotlib.cm.ScalarMappable(cmap=colormap)
        # pixmap = QtGui.QPixmap(32*2, 32)
        data = np.zeros((128 * 8, 128 * 16, 4), dtype=np.uint8)

        mi, ma = 1 * 10**self.mod1, self.data3d.max() * 10**self.mod2
        intensity_normalized = (np.log(self.data3d + 1.) - np.log(mi)) / (np.log(ma) - np.log(mi))
        import PIL.Image
        for y2d in range(8):
            for x2d in range(16):
                zindex = x2d + y2d * 16
                I = intensity_normalized[zindex]
                rgba = mapping.to_rgba(I, bytes=True)  # .reshape(Nx, 4)
                print(rgba.shape)
                subdata = data[y2d * 128:(y2d + 1) * 128, x2d * 128:(x2d + 1) * 128]
                for i in range(3):
                    subdata[:, :, i] = rgba[:, :, i]
                subdata[:, :, 3] = (intensity_normalized[zindex] * 255).astype(np.uint8)
                if 0:
                    filename = "cube%03d.png" % zindex
                    img = PIL.Image.frombuffer("RGB", (128, 128), subdata[:, :, 0:3] * 1)
                    print("saving to", filename)
                    img.save(filename)
        img = PIL.Image.frombuffer("RGBA", (128 * 16, 128 * 8), data)
        filename = "cube.png"
        print("saving to", filename)
        img.save(filename)

        filename = "colormap.png"
        print("saving to", filename)
        height, width = self.colormap_data.shape[:2]
        img = PIL.Image.frombuffer("RGB", (width, height), self.colormap_data)
        img.save(filename)


class TestWidget(QtGui.QMainWindow):
    def __init__(self, parent):
        super(TestWidget, self).__init__(parent)
        self.resize(700, 700)
        self.show()
        self.raise_()
        shortcut = QtGui.QShortcut(QtGui.QKeySequence("Cmd+Q"), self)
        shortcut.activated.connect(self.myclose)

        self.main = VolumeRenderWidget(self)
        self.setCentralWidget(self.main)

        # self.layout = QtGui.QVBoxLayout(self)
        # self.layout.addWidget(self.main)
        # self.setLayout(self.layout)

    def myclose(self, ignore=None):
        self.hide()


class Arrow(object):
    def begin(self, type):
        self.type = type

    def end(self):
        self.offset = len(self.vertices)

    def vertex3f(self, x, y, z):
        self.vertices.append([x, y, z])
        self.normals.append(list(self.current_normal))

    def normal3f(self, x, y, z):
        self.current_normal = [x, y, z]

    def tri(self, i1, i2, i3):
        self.indices.append(self.offset + i1)
        self.indices.append(self.offset + i2)
        self.indices.append(self.offset + i3)

    def drawGL(self, instances=1):
        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_NORMAL_ARRAY)
        vertices_ptr = self.vertices.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        glVertexPointer(3, GL_FLOAT, 0, vertices_ptr)
        normal_ptr = self.normals.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        glNormalPointer(GL_FLOAT, 0, normal_ptr)
        indices_ptr = self.indices.ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        # glDrawElements(GL_TRIANGLES, len(self.indices), GL_UNSIGNED_INT, indices_ptr);
        glDrawElementsInstanced(GL_TRIANGLES, len(self.indices), GL_UNSIGNED_INT, indices_ptr, instances)
        glDisableClientState(GL_VERTEX_ARRAY)
        glDisableClientState(GL_NORMAL_ARRAY)

    def __init__(self, x, y, z, scale=80, headfraction=0.4, baseradius=0.1, headradius=0.2):
        self.vertices = []
        self.normals = []
        self.indices = []
        self.offset = 0

        headfraction = 0.5
        baseradius = 0.1 * scale / 1.5
        headradius = 0.2 * scale

        # draw base
        self.begin(GL_QUADS)
        # glColor3f(1., 0., 0.)
        parts = 10
        for part in range(parts):
            angle = np.radians(part / 10. * 360)
            angle2 = np.radians((part + 1) / 10. * 360)
            self.normal3f(np.cos(angle), np.sin(angle), 0.)
            self.vertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z + scale / 2 - headfraction * scale)
            self.normal3f(np.cos(angle), np.sin(angle), 0.)
            self.vertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z - scale / 2)

            # self.normal3f(np.cos(angle2), np.sin(angle2), 0.)
            # self.vertex3f(x+baseradius*np.cos(angle2), y+baseradius*np.sin(angle2), z-scale/2)
            # self.normal3f(np.cos(angle2), np.sin(angle2), 0.)
            # self.vertex3f(x+baseradius*np.cos(angle2), y+baseradius*np.sin(angle2), z+scale/2-headfraction*scale)

        for part in range(10 + 1):
            self.tri((part * 2 + 0) % (10 * 2), (part * 2 + 1) % (10 * 2), (part * 2 + 2) % (10 * 2))
            self.tri((part * 2 + 2) % (10 * 2), (part * 2 + 1) % (10 * 2), (part * 2 + 3) % (10 * 2))
        self.end()

        # end of base
        self.begin(GL_TRIANGLE_FAN)
        self.normal3f(0, 0, -1)
        # glColor3f(0., 1., 0.)
        for part in range(10):
            angle = np.radians(-part / 10. * 360)
            self.vertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z - scale / 2)
        self.vertex3f(x, y, z - scale / 2)
        for part in range(parts + 1):
            self.tri(parts, part, (part + 1) % (parts))
        self.end()

        # glColor3f(0., 0., 1.)
        # head
        a = headradius - baseradius
        b = headfraction * scale
        headangle = np.arctan(a / b)
        for part in range(10 + 1):
            self.begin(GL_TRIANGLES)
            angle = np.radians(-part / 10. * 360)
            anglemid = np.radians(-(part + 0.5) / 10. * 360)
            angle2 = np.radians(-(part + 1) / 10. * 360)
            self.normal3f(np.cos(anglemid) * np.cos(headangle), np.sin(anglemid) * np.cos(headangle), np.sin(headangle))
            self.vertex3f(x, y, z + scale / 2)

            self.normal3f(np.cos(angle2) * np.cos(headangle), np.sin(angle2) * np.cos(headangle), np.sin(headangle))
            self.vertex3f(x + headradius * np.cos(angle2), y + headradius * np.sin(angle2), z + scale / 2 - headfraction * scale)

            self.normal3f(np.cos(angle) * np.cos(headangle), np.sin(angle) * np.cos(headangle), np.sin(headangle))
            self.vertex3f(x + headradius * np.cos(angle), y + headradius * np.sin(angle), z + scale / 2 - headfraction * scale)
            self.tri(0, 1, 2)
            self.end()

        # connecting base and head
        self.begin(GL_QUADS)
        # glColor3f(1., 1., 0.)
        self.normal3f(0, 0, -1)
        for part in range(10):
            angle = np.radians(-part / 10. * 360)
            angle2 = np.radians(-(part + 1) / 10. * 360)
            self.vertex3f(x + baseradius * np.cos(angle), y + baseradius * np.sin(angle), z + scale / 2 - headfraction * scale)
            self.vertex3f(x + headradius * np.cos(angle), y + headradius * np.sin(angle), z + scale / 2 - headfraction * scale)
            # self.vertex3f(x+baseradius*np.cos(angle2), y+baseradius*np.sin(angle2), z+scale/2-headfraction*scale)
            # self.vertex3f(x+headradius*np.cos(angle2), y+headradius*np.sin(angle2), z+scale/2-headfraction*scale)
        for part in range(10 + 1):
            self.tri((part * 2 + 0) % (10 * 2), (part * 2 + 1) % (10 * 2), (part * 2 + 2) % (10 * 2))
            self.tri((part * 2 + 2) % (10 * 2), (part * 2 + 1) % (10 * 2), (part * 2 + 3) % (10 * 2))
        self.end()

        self.vertices = np.array(self.vertices, dtype=np.float32)
        self.normals = np.array(self.normals, dtype=np.float32)
        self.indices = np.array(self.indices, dtype=np.uint32)


if __name__ == "__main__":
    # import vaex.ui.colormaps
    colormaps = vaex.ui.colormaps.colormaps
    import json
    js = json.dumps(vaex.ui.colormaps.colormaps)
    print(js)

    app = QtGui.QApplication(sys.argv)
    widget = TestWidget(None)

    def load():
        widget.main.loadTable(sys.argv[1], sys.argv[2:])
    widget.main.post_init = load
    sys.exit(app.exec_())
